"""VirusTotal v3 API."""

import asyncio
from enum import Enum
from typing import Any, Dict, List, Optional, Set, Tuple

import pandas as pd
from IPython.display import HTML, display

from ...common.exceptions import MsticpyImportExtraError
from ...common.provider_settings import get_provider_settings
from ...common.utility import is_ipython

try:
    import nest_asyncio
    import vt

    # Removing dependency temporarily due to build break
    from vt_graph_api import VTGraph
    from vt_graph_api import errors as vt_graph_errs

    from .vtfile_behavior import VTFileBehavior
except ImportError as imp_err:
    raise MsticpyImportExtraError(
        "Cannot use this feature without vt-py, vt-graph-api and",
        "nest_asyncio packages installed.",
        title="Error importing VirusTotal modules.",
        extra="vt3",
    ) from imp_err


# pylint: disable=too-many-lines


class MsticpyVTNoDataError(Exception):
    """No data returned from VT API."""


class MsticpyVTGraphSaveGraphError(Exception):
    """Could not save VT Graph."""


class VTEntityType(Enum):
    """VTEntityType: Enum class for VirusTotal entity types."""

    FILE = "file"
    DOMAIN = "domain"
    IP_ADDRESS = "ip_address"
    URL = "url"


class ColumnNames(Enum):
    """Column name enum for DataFrame output."""

    ID = "id"
    TYPE = "type"
    DETECTIONS = "detections"
    SCANS = "scans"
    SOURCE = "source"
    TARGET = "target"
    RELATIONSHIP_TYPE = "relationship_type"
    SOURCE_TYPE = "source_type"
    TARGET_TYPE = "target_type"


class VTObjectProperties(Enum):
    """Enum for VT Object properties."""

    ATTRIBUTES = "attributes"
    RELATIONSHIPS = "relationship"
    LAST_ANALYSIS_STATS = "last_analysis_stats"
    MALICIOUS = "malicious"


def _ensure_eventloop(force_nest_asyncio: bool = False):
    """Ensure that we have an event loop available."""
    try:
        event_loop = asyncio.get_event_loop()
    except RuntimeError:
        # Generate an event loop if there isn't any.
        event_loop = asyncio.new_event_loop()
        asyncio.set_event_loop(event_loop)
    if is_ipython() or force_nest_asyncio:
        nest_asyncio.apply()
    return event_loop


def _make_sync(future):
    """Wait for an async call, making it sync."""
    event_loop = _ensure_eventloop()
    return event_loop.run_until_complete(future)


VT_API_NOT_FOUND = "NotFoundError"


class VTLookupV3:
    """VTLookupV3: VirusTotal lookup of IoC reports."""

    _SUPPORTED_VT_TYPES: Set[VTEntityType] = {
        VTEntityType.FILE,
        VTEntityType.URL,
        VTEntityType.IP_ADDRESS,
        VTEntityType.DOMAIN,
    }

    _MAPPING_TYPES_ENDPOINT: Dict[VTEntityType, str] = {
        VTEntityType.FILE: "files",
        VTEntityType.URL: "urls",
        VTEntityType.IP_ADDRESS: "ip_addresses",
        VTEntityType.DOMAIN: "domains",
    }

    _BASIC_PROPERTIES_PER_TYPE: Dict[VTEntityType, Set[str]] = {
        VTEntityType.FILE: {
            "type_description",
            "size",
            "first_submission_date",
            "last_submission_date",
            "times_submitted",
            "meaningful_name",
        },
        VTEntityType.URL: {
            "first_submission_date",
            "last_submission_date",
            "times_submitted",
        },
        VTEntityType.IP_ADDRESS: {"date", "country", "asn", "as_owner"},
        VTEntityType.DOMAIN: {"id", "creation_date", "last_update_date", "country"},
    }

    _DEFAULT_SEARCH_LIMIT = 1000  # prevents vague queries from pulling 1M+ files
    _SEARCH_API_ENDPOINT = "/intelligence/search"

    @property
    def supported_vt_types(self) -> List[str]:
        """
        Return list of VirusTotal supported IoC type names.

        Returns
        -------
        List[str]:
            List of VirusTotal supported IoC type names.

        """
        return [str(i_type) for i_type in self._SUPPORTED_VT_TYPES]

    @classmethod
    def _get_endpoint_name(cls, vt_type: str) -> str:
        if VTEntityType(vt_type) not in cls._SUPPORTED_VT_TYPES:
            raise KeyError(f"Property type {vt_type} not supported")

        return cls._MAPPING_TYPES_ENDPOINT[VTEntityType(vt_type)]

    @classmethod
    def _parse_vt_object(
        cls, vt_object: vt.object.Object, all_props: bool = False
    ) -> pd.DataFrame:
        obj_dict = vt_object.to_dict()
        add_columns_df = pd.DataFrame()
        if VTObjectProperties.ATTRIBUTES.value in obj_dict:
            attributes = obj_dict[VTObjectProperties.ATTRIBUTES.value]
            vt_type = VTEntityType(vt_object.type)
            if vt_type not in cls._SUPPORTED_VT_TYPES:
                raise KeyError(f"Property type {vt_type} not supported")
            if not all_props:
                obj = {
                    key: attributes[key]
                    for key in cls._BASIC_PROPERTIES_PER_TYPE[vt_type]
                    if key in attributes
                }
            else:
                obj = attributes
            vt_df = pd.json_normalize(data=[obj])
            last_analysis_stats = attributes.get(
                VTObjectProperties.LAST_ANALYSIS_STATS.value
            )
            if last_analysis_stats:
                add_columns_df = pd.DataFrame(
                    {
                        ColumnNames.DETECTIONS.value: [
                            last_analysis_stats[VTObjectProperties.MALICIOUS.value]
                        ],
                        ColumnNames.SCANS.value: [sum(last_analysis_stats.values())],
                    }
                )
            # Format dates for pandas
            vt_df = timestamps_to_utcdate(vt_df)
        elif obj_dict:
            vt_df = pd.json_normalize([obj_dict])
        else:
            vt_df = cls._item_not_found_df(
                vt_type=vt_object.type, observable=vt_object.id
            )

        # Inject ID and Type columns
        if ColumnNames.ID.value not in vt_df.columns:
            add_columns_df[ColumnNames.ID.value] = [vt_object.id]
        if ColumnNames.TYPE.value not in vt_df.columns:
            add_columns_df[ColumnNames.TYPE.value] = [vt_object.type]
        return pd.concat(
            [vt_df, add_columns_df], axis="columns"
        )  # .set_index([ColumnNames.ID.value])

    def __init__(self, vt_key: Optional[str] = None, force_nestasyncio: bool = False):
        """
        Create a new instance of VTLookupV3 class.

        Parameters
        ----------
        vt_key: str, optional
            VirusTotal API key, if not supplied, this is read from
            user configuration.
        force_nestasyncio: bool, optional
            To force use of nest_asyncio, by default nest_asyncio
            is used in Jupyter notebooks, otherwise this defaults to False

        """
        _ensure_eventloop(force_nestasyncio)
        self._vt_key = vt_key or _get_vt_api_key()
        self._vt_client = vt.Client(apikey=self._vt_key)

    async def _lookup_ioc_async(
        self, observable: str, vt_type: str, all_props: bool = False
    ) -> pd.DataFrame:
        """
        Look up and single IoC observable.

        Parameters
        ----------
        observable: str
            The observable value
        vt_type: str
            The VT entity type
        all_props : bool, optional
            If True, return all properties, by default False

        Returns
        -------
            Attributes Pandas DataFrame with the properties of the entity

        Raises
        ------
        KeyError
            Unknown vt_type

        """
        if VTEntityType(vt_type) not in self._SUPPORTED_VT_TYPES:
            raise KeyError(
                f"Property type {vt_type} not supported",
                "Valid types are",
                ", ".join(x.value for x in VTEntityType.__members__.values()),
            )

        endpoint_name = self._get_endpoint_name(vt_type)
        try:
            response = self._vt_client.get_object(f"/{endpoint_name}/{observable}")
            return self._parse_vt_object(response, all_props=all_props)
        except vt.APIError as err:
            if err.args and err.args[0] == VT_API_NOT_FOUND:
                return self._item_not_found_df(vt_type, observable)
            raise MsticpyVTNoDataError(
                "An error occurred requesting data from VirusTotal"
            ) from err

    def lookup_ioc(
        self, observable: str, vt_type: str, all_props: bool = False
    ) -> pd.DataFrame:
        """
        Look up and single IoC observable.

        Parameters
        ----------
        observable: str
            The observable value
        vt_type: str
            The VT entity type
        all_props : bool, optional
            If True, return all properties, by default False

        Returns
        -------
            Attributes Pandas DataFrame with the properties of the entity

        Raises
        ------
        KeyError
            Unknown vt_type

        """
        try:
            return _make_sync(
                self._lookup_ioc_async(observable, vt_type, all_props=all_props)
            )
        finally:
            self._vt_client.close()

    async def _lookup_iocs_async(
        self,
        observables_df: pd.DataFrame,
        observable_column: str = ColumnNames.TARGET.value,
        observable_type_column: str = ColumnNames.TARGET_TYPE.value,
        all_props: bool = False,
    ):
        """
        Look up and multiple IoC observables.

        Parameters
        ----------
        observables_df: pd.DataFrame
            A Pandas DataFrame, where each row is an observable
        observable_column:
            ID column of each observable
        observable_type_column:
            Type column of each observable
        all_props : bool, optional
            If True, return all properties, by default False

        Returns
        -------
            Future Attributes Pandas DataFrame with the properties of the entities

        Raises
        ------
        KeyError
            Column not found in observables_df

        """
        _observables_df = observables_df.reset_index()

        for column in [observable_column, observable_type_column]:
            if column not in _observables_df.columns:
                raise KeyError(f"Column {column} not found in observables_df")

        observables_list = _observables_df[observable_column]
        types_list = _observables_df[observable_type_column]
        dfs_futures = []
        for observable, observable_type in zip(observables_list, types_list):
            try:
                ioc_df_future = self._lookup_ioc_async(
                    observable, observable_type, all_props=all_props
                )
                dfs_futures.append(ioc_df_future)
            except KeyError:
                print(
                    "ERROR\t It was not possible to obtain results for",
                    f"{observable_type} {observable}",
                )
        dfs = await asyncio.gather(*dfs_futures)
        return (
            pd.concat(dfs)
            if dfs
            else self._item_not_found_df(
                vt_type=", ".join(types_list),
                observable=", ".join(observables_list),
            )
        )

    def lookup_iocs(
        self,
        observables_df: pd.DataFrame,
        observable_column: str = ColumnNames.TARGET.value,
        observable_type_column: str = ColumnNames.TARGET_TYPE.value,
        all_props: bool = False,
    ):
        """
        Look up and multiple IoC observables.

        Parameters
        ----------
        observables_df: pd.DataFrame
            A Pandas DataFrame, where each row is an observable
        observable_column:
            ID column of each observable
        observable_type_column:
            Type column of each observable
        all_props : bool, optional
            If True, return all properties, by default False

        Returns
        -------
            Attributes Pandas DataFrame with the properties of the entities

        """
        try:
            return _make_sync(
                self._lookup_iocs_async(
                    observables_df,
                    observable_column,
                    observable_type_column,
                    all_props=all_props,
                )
            )
        finally:
            self._vt_client.close()

    # pylint: disable=too-many-locals
    async def _lookup_ioc_relationships_async(
        self,
        observable: str,
        vt_type: str,
        relationship: str,
        limit: Optional[int] = None,
        all_props: bool = False,
        full_objects: bool = False,
    ):
        """
        Look up and single IoC observable relationships.

        Parameters
        ----------
        observable: str
            The observable value
        vt_type: str
            The VT entity type
        relationship: str
            Desired relationship
        limit: int
            Relations limit
        all_props : bool, optional
            If True, return all properties, by default False
        full_objects : bool, optional
            If True, return the full object rather than just ID links.

        Returns
        -------
            Future Relationship Pandas DataFrame with the relationships of the entity

        Raises
        ------
        KeyError
            Unknown vt_type

        """
        if VTEntityType(vt_type) not in self._SUPPORTED_VT_TYPES:
            raise KeyError(f"Property type {vt_type} not supported")

        endpoint_name = self._get_endpoint_name(vt_type)
        response: vt.object.Object

        if limit is None:
            try:
                response = self._vt_client.get_object(
                    f"/{endpoint_name}/{observable}?relationship_counters=true"
                )
                relationships = response.relationships
                limit = (
                    relationships[relationship]["meta"]["count"]
                    if relationship in relationships
                    else 0
                )
            except KeyError:
                print(
                    f"ERROR: Could not obtain relationship limit for {vt_type} {observable}"
                )
                return self._item_not_found_df(vt_type=vt_type, observable=observable)

        if limit == 0 or limit is None:
            return self._item_not_found_df(vt_type=vt_type, observable=observable)

        obj_relation_type = (
            f"{relationship}" if full_objects else f"relationships/{relationship}"
        )
        try:
            # print(f"Obtaining {limit} relationships for {vt_type} {observable}")
            response = self._vt_client.iterator(
                f"/{endpoint_name}/{observable}/{obj_relation_type}",
                batch_size=40,
                limit=limit,
            )
            vt_objects = [
                self._parse_vt_object(r, all_props=all_props) for r in response
            ]
            result_df = (
                pd.concat(vt_objects, ignore_index=True)
                if vt_objects
                else self._relation_not_found_df(
                    vt_type=vt_type, observable=observable, relationship=relationship
                )
            )

            if vt_objects:
                # Add additional columns (with same size as DF)
                rows = len(result_df)
                add_columns = pd.DataFrame(
                    {
                        ColumnNames.SOURCE.value: [observable] * rows,
                        ColumnNames.SOURCE_TYPE.value: [VTEntityType(vt_type).value]
                        * rows,
                        ColumnNames.RELATIONSHIP_TYPE.value: [relationship] * rows,
                    }
                )

                result_df = (
                    pd.concat([result_df, add_columns], axis="columns")
                    .rename(
                        columns={
                            ColumnNames.ID.value: ColumnNames.TARGET.value,
                            ColumnNames.TYPE.value: ColumnNames.TARGET_TYPE.value,
                        },
                    )
                    .reset_index()
                )
                # result_df.set_index(
                #     [ColumnNames.SOURCE.value, ColumnNames.TARGET.value], inplace=True
                # )
                # if "index" in result_df.columns:
                #     result_df.drop(columns=["index"], inplace=True)
        except vt.APIError as err:
            if err.args and err.args[0] == VT_API_NOT_FOUND:
                return self._relation_not_found_df(vt_type, observable, relationship)
            raise MsticpyVTNoDataError(
                "An error occurred requesting data from VirusTotal"
            ) from err

        return result_df

    def lookup_ioc_relationships(
        self,
        observable: str,
        vt_type: str,
        relationship: str,
        limit: Optional[int] = None,
        all_props: bool = False,
    ) -> pd.DataFrame:
        """
        Look up single IoC observable relationship links.

        Parameters
        ----------
        observable: str
            The observable value
        vt_type: str
            The VT entity type
        relationship: str
            Desired relationship
        limit: int
            Relations limit
        all_props : bool, optional
            If True, return all properties, by default False

        Returns
        -------
            Relationship Pandas DataFrame with the relationships of the entity

        Notes
        -----
        This method returns relationship links rather than whole objects.
        That is, it will return the IDs of related items in the specified
        `relationship`, if any.

        See Also
        --------
        lookup_ioc_related : return the full related objects.

        """
        try:
            return _make_sync(
                self._lookup_ioc_relationships_async(
                    observable, vt_type, relationship, limit, all_props=all_props
                )
            )
        finally:
            self._vt_client.close()

    def lookup_ioc_related(
        self, observable: str, vt_type: str, relationship: str, limit: int = None
    ) -> pd.DataFrame:
        """
        Look single IoC observable related items.

        Parameters
        ----------
        observable: str
            The observable value
        vt_type: str
            The VT entity type
        relationship: str
            Desired relationship
        limit: int
            Relations limit

        Returns
        -------
        pd.DataFrame
            Any objects with specified `relationship` to the entity

        Notes
        -----
        This method returns full related objects rather than ID links.
        It is less efficient than looking up ID links only.

        See Also
        --------
        lookup_ioc_relationships : return the related IDs.

        """
        try:
            return _make_sync(
                self._lookup_ioc_relationships_async(
                    observable,
                    vt_type,
                    relationship,
                    limit,
                    all_props=True,
                    full_objects=True,
                )
            )
        finally:
            self._vt_client.close()

    async def _lookup_iocs_relationships_async(
        self,
        observables_df: pd.DataFrame,
        relationship: str,
        observable_column: str = ColumnNames.TARGET.value,
        observable_type_column: str = ColumnNames.TARGET_TYPE.value,
        limit: Optional[int] = None,
        all_props: bool = False,
    ) -> pd.DataFrame:
        """
        Look up and single IoC observable relationships.

        Parameters
        ----------
        observables_df: pd.DataFrame
            A Pandas DataFrame, where each row is an observable
        relationship: str
            Desired relationship
        observable_column:
            ID column of each observable
        observable_type_column:
            Type column of each observable.
        limit: int
            Relations limit
        all_props : bool, optional
            If True, return all properties, by default False

        Returns
        -------
            Future Relationship Pandas DataFrame with the relationships of each observable.

        Raises
        ------
        KeyError
            Column not found in observables_df

        """
        _observables_df = observables_df.reset_index()

        for column in [observable_column, observable_type_column]:
            if column not in _observables_df.columns:
                raise KeyError(f"Column {column} not found in observables df")

        observables_list = _observables_df[observable_column]
        types_list = _observables_df[observable_type_column]
        dfs_futures = []

        for observable, observable_type in zip(observables_list, types_list):
            try:
                dfs_futures.append(
                    self._lookup_ioc_relationships_async(
                        observable,
                        observable_type,
                        relationship,
                        limit,
                        all_props=all_props,
                    )
                )
            except KeyError:
                print(
                    "ERROR:\t It was not possible to get the data for",
                    f"{observable_type} {observable}",
                )
        dfs = await asyncio.gather(*dfs_futures)
        return (
            pd.concat(dfs)
            if len(dfs) > 0
            else self._relation_not_found_df(
                vt_type=", ".join(types_list),
                observable=", ".join(observables_list),
                relationship=relationship,
            )
        )

    def lookup_iocs_relationships(
        self,
        observables_df: pd.DataFrame,
        relationship: str,
        observable_column: str = ColumnNames.TARGET.value,
        observable_type_column: str = ColumnNames.TARGET_TYPE.value,
        limit: Optional[int] = None,
        all_props: bool = False,
    ) -> pd.DataFrame:
        """
        Look up and single IoC observable relationships.

        Parameters
        ----------
        observables_df: pd.DataFrame
            A Pandas DataFrame, where each row is an observable
        relationship: str
            Desired relationship
        observable_column:
            ID column of each observable
        observable_type_column:
            Type column of each observable.
        limit: int
            Relations limit
        all_props : bool, optional
            If True, return all properties, by default False

        Returns
        -------
            Relationship Pandas DataFrame with the relationships of each observable.

        """
        try:
            return _make_sync(
                self._lookup_iocs_relationships_async(
                    observables_df,
                    relationship,
                    observable_column,
                    observable_type_column,
                    limit,
                    all_props=all_props,
                )
            )

        finally:
            self._vt_client.close()

    def create_vt_graph(
        self, relationship_dfs: List[pd.DataFrame], name: str, private: bool
    ) -> str:
        """
        Create a VirusTotal Graph with a set of Relationship DataFrames.

        Parameters
        ----------
        relationship_dfs:
            List of Relationship DataFrames
        name:
            New graph name
        private
            Indicates if the Graph is private or not.

        Returns
        -------
            Graph ID

        Raises
        ------
            ValueError when private is not indicated.
            ValueError when there are no relationship DataFrames
            MsticpyVTGraphSaveGraphError when Graph can not be saved

        """
        if not isinstance(private, bool):
            raise ValueError("Please indicate if Graph is private or not")

        nodes, edges = self.relationships_to_graph(relationship_dfs)
        graph = VTGraph(self._vt_key, name=name, private=private)
        graph.add_nodes(nodes)

        for edge in edges:
            graph.add_link(**edge)

        try:
            graph.save_graph()
        except vt_graph_errs.SaveGraphError as graph_err:
            graph_mssg = (
                [
                    "Please check you have Private Graph premium feature enabled in",
                    "your subscription. It is possible to create public Graphs",
                    "with 'private=False' input argument",
                ]
                if private
                else [""]
            )

            raise MsticpyVTGraphSaveGraphError(
                "Could not save Graph.", *graph_mssg
            ) from graph_err

        return graph.graph_id

    def get_object(self, vt_id: str, vt_type: str) -> pd.DataFrame:
        """
        Return the full VT object as a DataFrame.

        Parameters
        ----------
        vt_id : str
            The ID of the object
        vt_type : str
            The type of object to query.

        Returns
        -------
        pd.DataFrame
            Single column DataFrame with attribute names as
            index and values as data column.

        Raises
        ------
        KeyError
            Unrecognized VT Type
        MsticpyVTNoDataError
            Error requesting data from VT.

        Notes
        -----
        This calls the underlying VT get_object API directly and
        returns all attributes for the object - hence a very wide
        DataFrame.

        """
        if VTEntityType(vt_type) not in self._SUPPORTED_VT_TYPES:
            # pylint: disable=no-member
            raise KeyError(
                f"Property type {vt_type} not supported",
                "Valid types are",
                ", ".join(x.value for x in VTEntityType.__members__.values()),
            )
            # pylint: enable=no-member

        endpoint_name = self._get_endpoint_name(vt_type)
        try:
            response: vt.object.Object = self._vt_client.get_object(
                f"/{endpoint_name}/{vt_id}"
            )
            result_df = pd.DataFrame(
                {
                    "id": [response.id],
                    "type": [response.type],
                }
            )
            attribs = pd.json_normalize(response.to_dict()["attributes"])
            result_df = pd.concat([result_df, attribs], axis=1)
            result_df["context_attributes"] = response.to_dict().get(
                "context_attributes"
            )
            return timestamps_to_utcdate(result_df)
        except vt.APIError as err:
            if err.args and err.args[0] == VT_API_NOT_FOUND:
                return self._item_not_found_df(vt_type, vt_id)
            raise MsticpyVTNoDataError(
                "An error occurred requesting data from VirusTotal"
            ) from err
        finally:
            self._vt_client.close()

    def get_file_behavior(
        self,
        file_id: Optional[str] = None,
        file_summary: Optional[Dict[str, Any]] = None,
        sandbox: Optional[str] = None,
    ) -> VTFileBehavior:
        """
        Return a VTFileBehavior object with file detonation results.

        Parameters
        ----------
        file_id : Optional[str], optional
            The ID of the file to look up, by default None
        file_summary : Optional[Dict[str, Any]], optional
            VT file summary object dictionary, by default None
        sandbox : str, optional
            Name of specific sandbox to retrieve, by default None
            If None, it will retrieve the behavior summary.

        Returns
        -------
        VTFileBehavior

        """
        vt_behavior = VTFileBehavior(
            self._vt_key, file_id=file_id, file_summary=file_summary
        )
        vt_behavior.get_file_behavior(sandbox=sandbox)
        return vt_behavior

    def search(
        self, query: str, limit: Optional[int] = _DEFAULT_SEARCH_LIMIT
    ) -> pd.DataFrame:
        """
        Return results of a VT search query as a DataFrame.

        Parameters
        ----------
        query : str
            Virus Total Intelligence Search string
        limit : int (default: 1,000)
            Number of intended results

        Returns
        -------
        pd.DataFrame
            Search query results.

        """
        # run virus total intelligence search using iterator
        try:
            response_itr = self._vt_client.iterator(
                self._SEARCH_API_ENDPOINT, params={"query": query}, limit=limit
            )
            response_list = list(map(lambda item: item.to_dict(), response_itr))
        except vt.APIError as api_err:
            raise MsticpyVTNoDataError(
                f"The provided query returned 0 results because of an APIError: {api_err}"
            ) from api_err

        if len(response_list) == 0:
            raise MsticpyVTNoDataError("The provided query returned 0 results")

        response_df = self._extract_response(response_list)
        return timestamps_to_utcdate(response_df)

    def iterator(
        self, path, *path_args, params=None, cursor=None, limit=None, batch_size=0
    ) -> vt.Iterator:
        """
        Return an iterator for the collection specified by the given path.

        The endpoint specified by path must return a collection of objects. An
        example of such an endpoint are /comments and /intelligence/search.

        SOURCE: https://github.com/VirusTotal/vt-py/blob/
        6bf4decb5bbd80bfc60e74ee3caa4c9073cea38c/vt/client.py

        Parameters
        ----------
        path : str
            Path to API endpoint returning a collection.
        path_args: dict
            A variable number of arguments that are put into any placeholders used in path.
        params: dict
            Additional parameters passed to the endpoint.
        cursor: str
            Cursor for resuming the iteration at the point it was left previously.
            A cursor can be obtained with Iterator.cursor(). This
            cursor is not the same one returned by the VirusTotal API.
        limit: int
            Maximum number of objects that will be returned by the iterator.
            If a limit is not provided the iterator continues until it reaches the
            last object in the collection.
        batch_size: int
            Maximum number of objects retrieved on each call to the endpoint.
            If not provided the server will decide how many objects to return.

        Returns
        -------
        vt.Iterator
            An instance of vt.Iterator

        """
        return self._vt_client.iterator(
            path, path_args, params, cursor, limit, batch_size
        )

    def _extract_response(self, response_list: list) -> pd.DataFrame:
        """
        Convert list of dictionaries from search() function to DataFrame.

        Parameters
        ----------
        response_list : list
            A list of dictionaries representing a virus total object

        Returns
        -------
        pd.DataFrame
            A DataFrame with the attributes of the virus total object

        """
        # loop to convert from list of vt.Object to pd.DataFrame
        response_rows = []
        for response_item in response_list:
            # flatten nested dictionary and append id, type values
            response_item_df = pd.json_normalize(response_item["attributes"])
            response_item_df["id"] = response_item["id"]
            response_item_df["type"] = response_item["type"]

            response_rows.append(response_item_df)

        response_df = pd.concat(response_rows, axis=0, ignore_index=True)
        return timestamps_to_utcdate(response_df)

    @staticmethod
    def relationships_to_graph(
        relationship_dfs: List[pd.DataFrame],
    ) -> Tuple[List[Dict[str, Any]], List[Dict[str, Any]]]:
        """
        Generate nodes and edges from relationships.

        Parameters
        ----------
        relationship_dfs : List[pd.DataFrame]
            List of relationship DataFrames

        Returns
        -------
        Tuple[List[Dict[str, Any]], List[Dict[str, Any]]]
            List of nodes (node_id, node_type)
            List of edges (source_node, target_node, connection_type)

        Raises
        ------
        ValueError
            If an empty list is supplied.

        """
        if not relationship_dfs:
            raise ValueError("There are no relationship DataFrames")

        # Set indexes to source-target so that pd.concat aligns on these indexes
        relationship_dfs = [
            rel_df.set_index([ColumnNames.SOURCE.value, ColumnNames.TARGET.value])
            for rel_df in relationship_dfs
        ]
        concatenated_df = pd.concat(relationship_dfs).reset_index()

        # Create nodes DF, with source and target
        sources_df = (
            concatenated_df.groupby(ColumnNames.SOURCE.value)[
                ColumnNames.SOURCE_TYPE.value
            ]
            .first()
            .reset_index()
            .rename(
                columns={
                    ColumnNames.SOURCE.value: ColumnNames.ID.value,
                    ColumnNames.SOURCE_TYPE.value: ColumnNames.TYPE.value,
                }
            )
        )

        target_df = (
            concatenated_df.groupby(ColumnNames.TARGET.value)[
                ColumnNames.TARGET_TYPE.value
            ]
            .first()
            .reset_index()
            .rename(
                columns={
                    ColumnNames.TARGET.value: ColumnNames.ID.value,
                    ColumnNames.TARGET_TYPE.value: ColumnNames.TYPE.value,
                }
            )
        )

        nodes_df = pd.concat([sources_df, target_df])
        nodes = [
            {
                "node_id": row[ColumnNames.ID.value],
                "node_type": row[ColumnNames.TYPE.value],
            }
            for row in nodes_df.itertuples(index=False)
        ]

        edges = [
            {
                "source_node": row[ColumnNames.SOURCE.value],
                "target_node": row[ColumnNames.TARGET.value],
                "connection_type": row[ColumnNames.RELATIONSHIP_TYPE.value],
            }
            for row in concatenated_df.itertuples(index=False)
        ]
        return nodes, edges

    @staticmethod
    def render_vt_graph(graph_id: str, width: int = 800, height: int = 600):
        """
        Display a VTGraph in a Jupyter Notebook.

        Parameters
        ----------
        graph_id:
            Graph ID
        width
            Graph width.
        height
            Graph height

        """
        display(
            HTML(
                f"""
              <iframe
                src="https://www.virustotal.com/graph/embed/{graph_id}"
                width="{width}"
                height="{height}">
              </iframe>

            """
            )
        )

    @classmethod
    def _item_not_found_df(cls, vt_type: str, observable: str):
        not_found_dict = {
            ColumnNames.ID.value: observable,
            ColumnNames.TYPE.value: vt_type,
        }
        vte_type = VTEntityType(vt_type)
        if vte_type not in cls._SUPPORTED_VT_TYPES:
            not_found_dict["status"] = "Unsupported type"
        else:
            not_found_dict.update(
                {key: "Not found" for key in cls._BASIC_PROPERTIES_PER_TYPE[vte_type]}
            )
        return pd.DataFrame([not_found_dict])

    @classmethod
    def _relation_not_found_df(cls, vt_type: str, observable: str, relationship: str):
        not_found_dict = {
            ColumnNames.SOURCE.value: observable,
            ColumnNames.SOURCE_TYPE.value: vt_type,
            ColumnNames.RELATIONSHIP_TYPE.value: relationship,
            ColumnNames.TARGET.value: "Not found",
            ColumnNames.TARGET_TYPE.value: "Not found",
        }
        return pd.DataFrame([not_found_dict])


def _get_vt_api_key() -> Optional[str]:
    """Retrieve the VT key from settings."""
    prov_settings = get_provider_settings("TIProviders")
    vt_settings = prov_settings.get("VirusTotal")
    if vt_settings:
        return vt_settings.args.get("AuthKey")
    return None


def timestamps_to_utcdate(data: pd.DataFrame):
    """Replace Unix timestamps in VT data with Py/pandas Timestamp."""
    columns = data.columns
    for date_col in (
        col for col in columns if isinstance(col, str) and col.endswith("_date")
    ):
        data = (
            data.assign(pd_data=pd.to_datetime(data[date_col], unit="s", utc=True))  # type: ignore
            .drop(columns=date_col)
            .rename(columns={"pd_data": date_col})
        )
    return data[columns]



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Deprecated - module process_tree.py has moved.

See :py:mod:`msticpy.vis.process_tree`
"""
import warnings

from .._version import VERSION

__version__ = VERSION
__author__ = "Pete Bryan"

# flake8: noqa: F403, F401
# pylint: disable=wildcard-import, unused-wildcard-import, unused-import
from ..transform.process_tree_utils import get_ancestors  # noqa F401
from ..transform.process_tree_utils import (
    get_children,
    get_descendents,
    get_parent,
    get_process,
    get_process_key,
    get_root,
    get_root_tree,
    get_roots,
    get_siblings,
    get_summary_info,
    get_tree_depth,
)
from ..vis.process_tree import *
from ..vis.process_tree import build_process_tree, infer_schema

WARN_MSSG = (
    "This module has moved to msticpy.vis.process_tree\n"
    "Please change your import to reflect this new location."
    "This will be removed in MSTICPy v2.2.0"
)
warnings.warn(WARN_MSSG, category=DeprecationWarning)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Logging global config."""
import logging
import os
import sys
from typing import NamedTuple, Optional, Union

from .._version import VERSION
from ..common.pkg_config import get_config

__version__ = VERSION
__author__ = "Ian Hellen"

_MP_LOG_FILE_ENV = "MSTICPYLOGFILE"
_MP_LOG_LEVEL_ENV = "MSTICPYLOGLEVEL"


class LoggingConfig(NamedTuple):
    """Logging configuration tuple."""

    log_file: Optional[str] = None
    log_level: int = logging.WARNING


def set_logging_level(log_level: Union[int, str]):
    """
    Set global logging level.

    Parameters
    ----------
    log_level : Union[int, str]
        Either an integer - one of logging.INFO, logging.WARNING,
        etc. or the name of a valid logging level.

    Raises
    ------
    ValueError
        If an invalid logging level is passed.

    See Also
    --------
    logging

    """
    logger = logging.getLogger("msticpy")
    log_level_int = logging.getLevelName(log_level or "WARNING")
    if not isinstance(log_level_int, int):
        raise ValueError(f"Invalid log level specified: {log_level}")
    logger.setLevel(log_level_int)


def setup_logging():
    """Initiate logging."""
    logging_config = _get_logging_config()
    # encoding param only supported in 3.9+
    params_39 = {"encoding": "utf-8"} if sys.version_info >= (3, 9) else {}
    logging.basicConfig(
        filename=logging_config.log_file,
        level=logging_config.log_level,
        format="%(asctime)s: %(levelname)s - %(message)s (%(module)s#%(lineno)d)",
        **params_39,
    )


def _get_logging_config():
    """Obtain logging configuration from environment or config."""
    log_file = os.environ.get(_MP_LOG_FILE_ENV)
    log_level = os.environ.get(_MP_LOG_LEVEL_ENV)

    log_file = log_file or get_config("Logging.FileName", None)
    log_level = log_level or get_config("Logging.LogLevel", None)
    log_level_int = logging.getLevelName(log_level or "WARNING")
    if not isinstance(log_level_int, int):
        log_level_int = logging.WARNING
    return LoggingConfig(
        log_file=log_file,
        log_level=log_level_int,
    )



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Lookup Status class."""
from enum import Enum
from typing import NamedTuple, Optional

from .._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


class SanitizedObservable(NamedTuple):
    """Encoded version of observable."""

    observable: Optional[str]
    status: Optional[str]


class LookupStatus(Enum):
    """Threat intelligence lookup status."""

    OK = 0
    NOT_SUPPORTED = 1
    BAD_FORMAT = 2
    QUERY_FAILED = 3
    NO_DATA = 4
    OTHER = 10



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
CrowdSec Provider.

Input can be a single IoC observable or a pandas DataFrame containing
multiple observables. Processing may require an API key and
processing performance may be limited to a specific number of
requests per minute for the account type that you have.

"""
from typing import Any, Dict, Tuple

from ..._version import VERSION
from ..http_provider import APILookupParams
from .ti_http_provider import HttpTIProvider
from .ti_provider_base import ResultSeverity

__version__ = VERSION
__author__ = "Shivam Sandbhor"


class CrowdSec(HttpTIProvider):
    """CrowdSec CTI Smoke Lookup."""

    _BASE_URL = "https://cti.api.crowdsec.net"

    _QUERIES = {
        "ipv4": APILookupParams(
            path="/v2/smoke/{observable}",
            headers={
                "x-api-key": "{AuthKey}",
                "User-Agent": "crowdsec-msticpy-tiprovider/v1.0.0",
            },
        ),
    }
    _QUERIES["ipv6"] = _QUERIES["ipv4"]

    def parse_results(self, response: Dict) -> Tuple[bool, ResultSeverity, Any]:
        """Return the details of the response."""
        if self._failed_response(response):
            return False, ResultSeverity.information, response["RawResult"]["message"]

        if response["RawResult"]["scores"]["overall"]["total"] <= 2:
            result_severity = ResultSeverity.information
        elif response["RawResult"]["scores"]["overall"]["total"] <= 3:
            result_severity = ResultSeverity.warning
        else:
            result_severity = ResultSeverity.high

        return (
            True,
            result_severity,
            {
                "Background Noise": response["RawResult"]["background_noise_score"],
                "Overall Score": response["RawResult"]["scores"]["overall"]["total"],
                "First Seen": response["RawResult"]["history"]["first_seen"],
                "Last Seen": response["RawResult"]["history"]["last_seen"],
                "Attack Details": ",".join(
                    [
                        attack_detail["label"]
                        for attack_detail in response["RawResult"]["attack_details"]
                    ]
                ),
                "Behaviors": ",".join(
                    [
                        behavior["name"]
                        for behavior in response["RawResult"]["behaviors"]
                    ]
                ),
            },
        )



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module for pre-defined widget layouts."""
from typing import Any, Dict, List, Union

import ipywidgets as widgets

from .._version import VERSION
from .core import IPyDisplayMixin

__version__ = VERSION
__author__ = "Ian Hellen"


# pylint: disable=too-many-instance-attributes
class SelectSubset(IPyDisplayMixin):
    """Class to select a subset from an input list."""

    def __init__(
        self,
        source_items: Union[Dict[str, str], List[Any]],
        default_selected: Union[Dict[str, str], List[Any]] = None,
        display_filter: bool = True,
        auto_display: bool = True,
    ):
        """
        Create instance of SelectSubset widget.

        Parameters
        ----------
        source_items : Union[Dict[str, str], List[Any]]
            List of source items - either a dictionary(label, value),
            a simple list or
            a list of (label, value) tuples.
        default_selected : Union[Dict[str, str], List[Any]]
            Populate the selected list with values - either
            a dictionary(label, value),
            a simple list or
            a list of (label, value) tuples.
        display_filter : bool, optional
            Whether to display item filter (the default is True)
        auto_display : bool, optional
            Whether to display on instantiation (the default is True)

        """
        if isinstance(source_items, dict):
            source_items = list(source_items.items())

        self.src_items = sorted(set(source_items))
        if isinstance(self.src_items[0], tuple):
            self._src_dict = {val: (label, val) for label, val in self.src_items}
        else:
            self._src_dict = {}

        w_layout = widgets.Layout(width="40%", height="200px")
        self._source_list = widgets.SelectMultiple(
            options=sorted(set(self.src_items)), layout=w_layout, description="Source: "
        )

        if isinstance(default_selected, dict):
            default_selected = list(default_selected.items())
        if default_selected:
            set_selected = set(default_selected)
            selected_items = sorted(set_selected.intersection(source_items))
        else:
            selected_items = []

        self._select_list = widgets.SelectMultiple(
            options=selected_items, layout=w_layout, description="Selected: "
        )

        self._display_filter = display_filter
        if display_filter:
            self._w_filter = widgets.Text(
                value="", description="Filter:", style={"description_width": "initial"}
            )

            # set up observer callbacks
            self._w_filter.observe(self._update_options, names="value")

        self._b_add_all = widgets.Button(description="Add All \u21fe")
        self._b_add = widgets.Button(description="Add \u21fe")
        self._b_del = widgets.Button(description="\u21fd Remove")
        self._b_del_all = widgets.Button(description="\u21fd Remove All")

        self._b_add.on_click(self._on_btn_add)
        self._b_del.on_click(self._on_btn_del)
        self._b_del_all.on_click(self._on_btn_del_all)
        self._b_add_all.on_click(self._on_btn_add_all)

        v_box = widgets.VBox(
            [self._b_add_all, self._b_add, self._b_del, self._b_del_all]
        )
        self.layout = widgets.HBox([self._source_list, v_box, self._select_list])
        if self._display_filter:
            self.layout = widgets.VBox([self._w_filter, self.layout])
        if auto_display:
            self.display()

    @property
    def value(self) -> List[Any]:
        """Return currently selected value or values."""
        return self.selected_values

    @property
    def selected_items(self) -> List[Any]:
        """
        Return a list of the selected items.

        If the input list is a list of tuples, this returns
        a list of the selected tuples.

        Returns
        -------
        List[Any]
            List of items in the selected list.

        """
        return list(self._select_list.options)

    @property
    def selected_values(self) -> List[Any]:
        """
        Return list of selected values.

        If the input list is a list of tuples, this returns
        a list of values of the items.

        Returns
        -------
        List[Any]
            List of selected item values.

        """
        if self._select_list.options and isinstance(
            self._select_list.options[0], tuple
        ):
            return [item[1] for item in self._select_list.options]
        return self.selected_items

    def _update_options(self, change):
        """Filter the alert list by substring."""
        if change is not None and "new" in change:
            self._source_list.options = sorted(
                {
                    i
                    for i in self.src_items
                    if str(change["new"]).lower() in str(i).lower()
                }
            )

    # pylint: disable=not-an-iterable
    def _on_btn_add(self, button):
        del button
        selected_set = set(self._select_list.options)
        for selected in self._source_list.value:
            if self._src_dict:
                selected_set.add(self._src_dict[selected])
            else:
                selected_set.add(selected)
        self._select_list.options = sorted(list(selected_set))

    def _on_btn_add_all(self, button):
        del button
        self._select_list.options = sorted(list(set(self._source_list.options)))

    def _on_btn_del(self, button):
        del button
        selected_set = set(self._select_list.options)
        # save the current index
        cur_index = max(self._select_list.index)
        if selected_set:
            for selected in self._select_list.value:
                if self._src_dict:
                    selected_set.remove(self._src_dict[selected])
                else:
                    selected_set.remove(selected)
            self._select_list.options = sorted(list(selected_set))
        if not self._select_list.options:
            return
        # try to set the index to the next item in the list
        if cur_index < len(self._select_list.options):
            next_item = cur_index or 0
            self._select_list.index = tuple([next_item])
        else:
            last_item = max(len(self._select_list.options) - 1, 0)
            self._select_list.index = tuple([last_item])

    # pylint: enable=not-an-iterable

    def _on_btn_del_all(self, button):
        del button
        self._select_list.options = []



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Azure authentication handling."""
import os
from typing import List, Optional

from azure.common.exceptions import CloudError
from azure.identity import DeviceCodeCredential
from azure.mgmt.subscription import SubscriptionClient

from .._version import VERSION

# pylint: enable=unused-import
from ..common.provider_settings import get_provider_settings

# importing only_interactive_cred for client use.
# pylint: disable=unused-import
from .azure_auth_core import (  # noqa: F401
    AzCredentials,
    AzureCloudConfig,
    AzureCredEnvNames,
    az_connect_core,
    list_auth_methods,
    only_interactive_cred,
)
from .cred_wrapper import CredentialWrapper

__version__ = VERSION
__author__ = "Pete Bryan"


def az_connect(
    auth_methods: Optional[List[str]] = None,
    tenant_id: Optional[str] = None,
    silent: bool = False,
    **kwargs,
) -> AzCredentials:
    """
    Connect to Azure SDK/API.

    Parameters
    ----------
    auth_methods : List[str], optional
        List of authentication methods to try
        For a list of possible authentication methods use the `list_auth_methods`
        function.
        Default is ["env", "msi", "vscode", "cli", "powershell", "devicecode"]
    tenant_id : str, optional
        The tenant to authenticate against. If not supplied, the default
        tenant for the identity will be used.
    silent : bool, optional
        Set True to hide all output during connection, by default False
    credential : AzureCredential
        If an Azure credential is passed, it will be used directly.
    cloud : str, optional
        What Azure cloud to connect to.
        By default it will attempt to use the cloud setting from config file.
        If this is not set it will default to Azure Public Cloud

    Returns
    -------
    AzCredentials
        Named tuple of:
        - legacy (ADAL) credentials
        - modern (MSAL) credentials


    Raises
    ------
    CloudError
        If chained token credential creation fails.

    See Also
    --------
    list_auth_methods

    """
    az_cloud_config = AzureCloudConfig(cloud=kwargs.get("cloud"))
    # Use auth_methods param or configuration defaults
    data_provs = get_provider_settings(config_section="DataProviders")
    auth_methods = auth_methods or az_cloud_config.auth_methods
    tenant_id = tenant_id or az_cloud_config.tenant_id

    # Ignore AzCLI settings except for authentication creds for EnvCred
    az_cli_config = data_provs.get("AzureCLI")
    if (
        az_cli_config
        and az_cli_config.args
        and isinstance(auth_methods, list)
        and "env" in auth_methods
    ):
        os.environ[AzureCredEnvNames.AZURE_CLIENT_ID] = (
            az_cli_config.args.get("clientId") or ""
        )
        os.environ[AzureCredEnvNames.AZURE_TENANT_ID] = (
            az_cli_config.args.get("tenantId") or ""
        )
        os.environ[AzureCredEnvNames.AZURE_CLIENT_SECRET] = (
            az_cli_config.args.get("clientSecret") or ""
        )
    credentials = az_connect_core(
        auth_methods=auth_methods, tenant_id=tenant_id, silent=silent, **kwargs
    )
    sub_client = SubscriptionClient(
        credential=credentials.modern,
        base_url=az_cloud_config.resource_manager,  # type: ignore
        credential_scopes=[az_cloud_config.token_uri],
    )
    if not sub_client:
        raise CloudError("Could not create a Subscription client.")

    return credentials


def az_user_connect(
    tenant_id: Optional[str] = None, silent: bool = False
) -> AzCredentials:
    """
    Authenticate to the SDK using user based authentication methods, Azure CLI or interactive logon.

    Parameters
    ----------
    tenant_id : str, optional
        The tenant to authenticate against. If not supplied, the default tenant for the
        identity will be used.
    silent : bool, optional
        Whether you want the auth process to display any output, by default False

    Returns
    -------
    AzCredentials

    """
    return az_connect_core(
        auth_methods=["cli", "interactive"], tenant_id=tenant_id, silent=silent
    )


def fallback_devicecode_creds(
    cloud: Optional[str] = None, tenant_id: Optional[str] = None, **kwargs
):
    """
    Authenticate using device code as a fallback method.

    Parameters
    ----------
    cloud : str, optional
        What Azure cloud to connect to.
        By default it will attempt to use the cloud setting from config file.
        If this is not set it will default to Azure Public Cloud
    tenant_id : str, optional
        The tenant to authenticate against. If not supplied,
        the tenant ID is read from configuration, or the default tenant for the identity.

    Returns
    -------
    AzCredentials
                Named tuple of:
        - legacy (ADAL) credentials
        - modern (MSAL) credentials

    Raises
    ------
    CloudError
        If chained token credential creation fails.

    """
    cloud = cloud or kwargs.pop("region", AzureCloudConfig().cloud)
    az_config = AzureCloudConfig(cloud)
    aad_uri = az_config.authority_uri
    tenant_id = tenant_id or az_config.tenant_id
    creds = DeviceCodeCredential(authority=aad_uri, tenant_id=tenant_id)
    legacy_creds = CredentialWrapper(creds, resource_id=az_config.token_uri)
    if not creds:
        raise CloudError("Could not obtain credentials.")

    return AzCredentials(legacy_creds, creds)


def get_default_resource_name(resource_uri: str) -> str:
    """Get a default resource name for a resource URI."""
    separator = "" if resource_uri.strip().endswith("/") else "/"
    return f"{resource_uri}{separator}.default"



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Display code with with highlighting."""
from typing import List, Optional

from IPython.display import HTML, DisplayHandle, display
from pygments import formatters, highlight, lexers, styles

from .._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"


def to_html(code: str, language: str, style: str = "default", full: bool = True) -> str:
    """
    Return Pygments-highlighted code for specified language.

    Parameters
    ----------
    code : str
        Input code as string.
    language : str
        The language name or alias.
    style : str, optional
        The pygments style to use, by default "default"
    full : bool, optional
        By default this function will return a complete HTML document,
        to only return the formatted code snippet (minus styles), set
        full=False.

    Returns
    -------
    str
       HTML document with pygments formatting.

    See Also
    --------
    list_pygments_styles - list the available styles use

    """
    pygments_lexer = lexers.get_lexer_by_name(language)
    # pylint: disable=no-member
    return highlight(
        code,
        pygments_lexer,
        formatters.HtmlFormatter(style=style, full=full, nobackground=True),
    )
    # pylint: enable=no-member


def list_pygments_styles() -> List[str]:
    """
    Return list of pygments styles available.

    Returns
    -------
    List[str]
        The list of pygments style names.

    """
    return list(styles.STYLE_MAP)


def display_html(
    code: str,
    language: str,
    style: str = "stata-dark",
    full: bool = True,
    display_handle: bool = False,
) -> Optional[DisplayHandle]:
    """
    Display pygments-formatted code.

    Parameters
    ----------
    code : str
        Input code as string.
    language : str
        The language name or alias.
    style : str, optional
        The pygments style to use, by default "stata-dark"
    full : bool, optional
        By default this function will return a complete HTML document,
        to only return the formatted code snippet (minus styles), set
        full=False.
    display_handle : bool, optional
        If True, will return an IPython DisplayHandle for the display
        object created.

    Returns
    -------
    Optional[DisplayHandle] :
        Handle to the IPython display object.

    See Also
    --------
    to_html

    """
    html = to_html(code, language, style, full)
    if display_handle:
        return display(HTML(html), display_id=True)
    display(HTML(html))
    return None



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Module for TILookup classes.

Input can be a single IoC observable or a pandas DataFrame containing
multiple observables. Processing may require a an API key and
processing performance may be limited to a specific number of
requests per minute for the account type that you have.

"""

from typing import Dict, Iterable, List, Mapping, Optional, Union

import pandas as pd

from .._version import VERSION
from ..common.utility import export
from .lookup import Lookup

# used in dynamic instantiation of providers
from .provider_base import Provider, _make_sync
from .tiproviders import TI_PROVIDERS

__version__ = VERSION
__author__ = "Ian Hellen"


@export
class TILookup(Lookup):
    """Threat Intel observable lookup from providers."""

    _NO_PROVIDERS_MSG = """
    No TI Providers are loaded - please check that
    you have correctly configured your msticpyconfig.yaml settings.
    """
    _HELP_URI = (
        "https://msticpy.readthedocs.io/en/latest/data_acquisition/"
        "TIProviders.html#configuration-file"
    )

    PROVIDERS = TI_PROVIDERS
    PACKAGE = "tiproviders"
    CUSTOM_PROVIDERS: Dict[str, Provider] = {}

    # pylint: disable=too-many-arguments
    def lookup_ioc(
        self,
        ioc: Optional[str] = None,
        ioc_type: Optional[str] = None,
        ioc_query_type: Optional[str] = None,
        providers: Optional[List[str]] = None,
        default_providers: Optional[List[str]] = None,
        prov_scope: str = "primary",
        **kwargs,
    ) -> pd.DataFrame:
        """
        Lookup Threat Intelligence reports for a single IoC in active providers.

        Parameters
        ----------
        ioc : str
            IoC observable
        observable : str
            alias for `ioc`
        ioc_type : str, optional
            One of IoCExtract.IoCType, by default None
            If none, the IoC type will be inferred
        ioc_query_type: str, optional
            The ioc query type (e.g. rep, info, malware)
        providers: List[str]
            Explicit list of providers to use
        default_providers: Optional[List[str]], optional
            Used by pivot functions as a fallback to `providers`. If
            `providers` is specified, it will override this parameter.
        prov_scope : str, optional
            Use "primary", "secondary" or "all" providers, by default "primary"
        kwargs :
            Additional arguments passed to the underlying provider(s)

        Returns
        -------
        pd.DataFrame
            The result returned as a tuple(bool, list):
            bool indicates whether a TI record was found in any provider
            list has an entry for each provider result

        See Also
        --------
        lookup_iocs : Lookup Threat Intelligence reports for a collection of IoCs.

        Notes
        -----
        Queries active Threat Intelligence (TI) providers for a single
        indicator of compromise (IoC). It returns results as a pandas
        DataFrame. `ioc_type` can be used to specify the type (ipv4,
        ipv6, dns, url, file_hash). If this is not supplied the
        type is inferred using regular expressions.
        By default, providers are queried asynchronously, in parallel.

        """
        ioc = ioc or kwargs.pop("observable", None)
        if ioc is None:
            raise ValueError("No value supplied for 'ioc' parameter")
        return self.lookup_item(
            item=ioc,
            item_type=ioc_type,
            query_type=ioc_query_type,
            providers=providers,
            default_providers=default_providers,
            prov_scope=prov_scope,
            **kwargs,
        )

    def lookup_iocs(
        self,
        data: Union[pd.DataFrame, Mapping[str, str], Iterable[str]],
        ioc_col: str = None,
        ioc_type_col: str = None,
        ioc_query_type: str = None,
        providers: List[str] = None,
        default_providers: Optional[List[str]] = None,
        prov_scope: str = "primary",
        **kwargs,
    ) -> pd.DataFrame:
        """
        Lookup Threat Intelligence reports for a collection of IoCs in active providers.

        Parameters
        ----------
        data : Union[pd.DataFrame, Mapping[str, str], Iterable[str]]
            Data input in one of three formats:
            1. Pandas dataframe (you must supply the column name in
            `ioc_col` parameter)
            2. Mapping (e.g. a dict) of [observable, IoCType]
            3. Iterable of observables - IoCTypes will be inferred
        ioc_col : str, optional
            DataFrame column to use for observables, by default None
            ("col" and "column" are also aliases for this parameter)
        ioc_type_col : str, optional
            DataFrame column to use for IoCTypes, by default None
        ioc_query_type: str, optional
            The ioc query type (e.g. rep, info, malware)
        providers: List[str]
            Explicit list of providers to use
        default_providers: Optional[List[str]], optional
            Used by pivot functions as a fallback to `providers`. If
            `providers` is specified, it will override this parameter.
        prov_scope : str, optional
            Use "primary", "secondary" or "all" providers, by default "primary"
        kwargs :
            Additional arguments passed to the underlying provider(s)

        Returns
        -------
        pd.DataFrame
            DataFrame of results

        See Also
        --------
        lookup_ioc : Lookup Threat Intelligence reports for a single IoC.

        Notes
        -----
        `lookup_iocs` queries active Threat Intelligence (TI) providers for
        threat reports. It can accept input as a Python iterable or
        a pandas dataframe. In the latter case, you also need to supply
        the `ioc_col` parameter to indicate which column the IoC value can
        be found. The `ioc_type_col` parameter is optional and can be used
        to manually specify the IoC type for each row. If this is not supplied
        the ioc types are inferred using regular expressions.
        The results are returned as a pandas DataFrame.

        """
        return _make_sync(
            self._lookup_iocs_async(
                data=data,
                ioc_col=ioc_col,
                ioc_type_col=ioc_type_col,
                ioc_query_type=ioc_query_type,
                providers=providers,
                default_providers=default_providers,
                prov_scope=prov_scope,
                **kwargs,
            )
        )

    # pylint: disable=too-many-locals
    async def _lookup_iocs_async(
        self,
        data: Union[pd.DataFrame, Mapping[str, str], Iterable[str]],
        ioc_col: str = None,
        ioc_type_col: str = None,
        ioc_query_type: str = None,
        providers: List[str] = None,
        default_providers: Optional[List[str]] = None,
        prov_scope: str = "primary",
        **kwargs,
    ) -> pd.DataFrame:
        """Lookup IoCs async."""
        return await self._lookup_items_async(
            data,
            item_col=ioc_col,
            item_type_col=ioc_type_col,
            ioc_query_type=ioc_query_type,
            providers=providers,
            default_providers=default_providers,
            prov_scope=prov_scope,
            **kwargs,
        )

    def lookup_iocs_sync(
        self,
        data: Union[pd.DataFrame, Mapping[str, str], Iterable[str]],
        ioc_col: str = None,
        ioc_type_col: str = None,
        ioc_query_type: str = None,
        providers: List[str] = None,
        default_providers: Optional[List[str]] = None,
        prov_scope: str = "primary",
        **kwargs,
    ) -> pd.DataFrame:
        """
        Lookup a collection of IoCs.

        Parameters
        ----------
        data : Union[pd.DataFrame, Mapping[str, str], Iterable[str]]
            Data input in one of three formats:
            1. Pandas dataframe (you must supply the column name in
            `ioc_col` parameter)
            2. Mapping (e.g. a dict) of [observable, IoCType]
            3. Iterable of observables - IoCTypes will be inferred
        ioc_col : str, optional
            DataFrame column to use for observables, by default None
            ("col" and "column" are also aliases for this parameter)
        ioc_type_col : str, optional
            DataFrame column to use for IoCTypes, by default None
        ioc_query_type: str, optional
            The ioc query type (e.g. rep, info, malware)
        providers: List[str]
            Explicit list of providers to use
        default_providers: Optional[List[str]], optional
            Used by pivot functions as a fallback to `providers`. If
            `providers` is specified, it will override this parameter.
        prov_scope : str, optional
            Use "primary", "secondary" or "all" providers, by default "primary"
        kwargs :
            Additional arguments passed to the underlying provider(s)

        Returns
        -------
        pd.DataFrame
            DataFrame of results

        """
        return self.lookup_items_sync(
            data,
            item_col=ioc_col,
            item_type_col=ioc_type_col,
            query_type=ioc_query_type,
            providers=providers,
            default_providers=default_providers,
            prov_scope=prov_scope,
            **kwargs,
        )

    def _load_providers(self, **kwargs):
        """Load provider classes based on config."""
        return super()._load_providers(providers="TIProviders", **kwargs)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module docstring."""
from typing import Iterable, Optional, Tuple, Union

import pandas as pd
from bokeh.models import LayoutDOM
from bokeh.plotting import figure

from .._version import VERSION
from ..common.exceptions import MsticpyUserError
from ..transform.network import GraphType, df_to_networkx
from ..vis.network_plot import plot_nx_graph
from .entity_graph_tools import EntityGraph, req_alert_cols, req_inc_cols
from .foliummap import plot_map
from .matrix_plot import plot_matrix
from .process_tree import build_and_show_process_tree
from .timeline import display_timeline
from .timeline_duration import display_timeline_duration
from .timeline_values import display_timeline_values

__version__ = VERSION
__author__ = "Ian Hellen"


@pd.api.extensions.register_dataframe_accessor("mp_plot")
class MsticpyPlotAccessor:
    """Pandas api extension for MSTICPy visualizations."""

    def __init__(self, pandas_obj):
        """Instantiate pandas extension class."""
        self._df = pandas_obj

    def timeline(self, **kwargs) -> LayoutDOM:
        """
        Display a timeline of events.

        Parameters
        ----------
        time_column : str, optional
            Name of the timestamp column
            (the default is 'TimeGenerated')
        source_columns : list, optional
            List of default source columns to use in tooltips
            (the default is None)

        Other Parameters
        ----------------
        title : str, optional
            Title to display (the default is None)
        alert : SecurityAlert, optional
            Add a reference line/label using the alert time (the default is None)
        ref_event : Any, optional
            Add a reference line/label using the alert time (the default is None)
        ref_time : datetime, optional
            Add a reference line/label using `ref_time` (the default is None)
        group_by : str
            The column to group timelines on.
        legend: str, optional
            "left", "right", "inline" or "none"
            (the default is to show a legend when plotting multiple series
            and not to show one when plotting a single series)
        yaxis : bool, optional
            Whether to show the yaxis and labels (default is False)
        ygrid : bool, optional
            Whether to show the yaxis grid (default is False)
        xgrid : bool, optional
            Whether to show the xaxis grid (default is True)
        range_tool : bool, optional
            Show the the range slider tool (default is True)
        height : int, optional
            The height of the plot figure
            (the default is auto-calculated height)
        width : int, optional
            The width of the plot figure (the default is 900)
        color : str
            Default series color (default is "navy")
        overlay_data : pd.DataFrame:
            A second dataframe to plot as a different series.
        overlay_color : str
            Overlay series color (default is "green")
        ref_events : pd.DataFrame, optional
            Add references line/label using the event times in the dataframe.
            (the default is None)
        ref_time_col : str, optional
            Add references line/label using the this column in `ref_events`
            for the time value (x-axis).
            (this defaults the value of the `time_column` parameter or 'TimeGenerated'
            `time_column` is None)
        ref_col : str, optional
            The column name to use for the label from `ref_events`
            (the default is None)
        ref_times : List[Tuple[datetime, str]], optional
            Add one or more reference line/label using (the default is None)

        Returns
        -------
        LayoutDOM
            The bokeh plot figure.

        """
        return display_timeline(data=self._df, **kwargs)

    # pylint: disable=invalid-name
    def timeline_values(self, value_column: str = None, **kwargs) -> LayoutDOM:
        """
        Display a timeline of events.

        Parameters
        ----------
        time_column : str, optional
            Name of the timestamp column
            (the default is 'TimeGenerated')
        value_column : str
            The column name holding the value to plot vertically
        source_columns : list, optional
            List of default source columns to use in tooltips
            (the default is None)

        Other Parameters
        ----------------
        x : str, optional
            alias of `time_column`
        y : str, optional
            alias of `value_column`
        value_col : str, optional
            alias of `value_column`
        title : str, optional
            Title to display (the default is None)
        ref_event : Any, optional
            Add a reference line/label using the alert time (the default is None)
        ref_time : datetime, optional
            Add a reference line/label using `ref_time` (the default is None)
        ref_label : str, optional
            A label for the `ref_event` or `ref_time` reference item
        group_by : str
            (where `data` is a DataFrame)
            The column to group timelines on
        legend: str, optional
            "left", "right", "inline" or "none"
            (the default is to show a legend when plotting multiple series
            and not to show one when plotting a single series)
        yaxis : bool, optional
            Whether to show the yaxis and labels
        range_tool : bool, optional
            Show the the range slider tool (default is True)
        height : int, optional
            The height of the plot figure
            (the default is auto-calculated height)
        width : int, optional
            The width of the plot figure (the default is 900)
        color : str
            Default series color (default is "navy"). This is overridden by
            automatic color assignments if plotting a grouped chart
        kind : Union[str, List[str]]
            one or more glyph types to plot., optional
            Supported types are "circle", "line" and "vbar" (default is "vbar")
        ref_events : pd.DataFrame, optional
            Add references line/label using the event times in the dataframe.
            (the default is None)
        ref_time_col : str, optional
            Add references line/label using the this column in `ref_events`
            for the time value (x-axis).
            (this defaults the value of the `time_column` parameter or 'TimeGenerated'
            `time_column` is None)
        ref_col : str, optional
            The column name to use for the label from `ref_events`
            (the default is None)
        ref_times : List[Tuple[datetime, str]], optional
            Add one or more reference line/label using (the default is None)

        Returns
        -------
        LayoutDOM
            The bokeh plot figure.

        """
        return display_timeline_values(
            data=self._df, value_column=value_column, **kwargs
        )

    def timeline_duration(
        self,
        group_by: Union[Iterable[str], str],
        time_column: str = "TimeGenerated",
        end_time_column: Optional[str] = None,
        **kwargs,
    ) -> LayoutDOM:  # noqa: C901, MC0001
        """
        Display a duration timeline of events grouped by one or more columns.

        Parameters
        ----------
        group_by : Union[Iterable[str], str]
            The column name or iterable of column names to group the data by.
        time_column : str
            Primary time column - will be used to calculate the
            start time of the duration for each group.
            If `end_time_column` is not specified it will also be used to
            calculate the end time.
        end_time_column : Optional[str]
            If supplied, it will be used to calculate the end time
            of the duration for each group.

        Other Parameters
        ----------------
        title : str, optional
            Title to display (the default is None)
        ylabel_cols : Optional[Iterable[str]], optional
            The subset of the group columns to use for the y-axis labels.
        yaxis : bool, optional
            Whether to show the yaxis and labels
        range_tool : bool, optional
            Show the the range slider tool (default is True)
        source_columns : list, optional
            List of default source columns to use in tooltips
            (the default is None)
        height : int, optional
            The height of the plot figure
            (the default is auto-calculated height)
        width : int, optional
            The width of the plot figure (the default is 900)
        color : str
            Default series color (default is "navy")
        ref_events : pd.DataFrame, optional
            Add references line/label using the event times in the dataframe.
            (the default is None)
        ref_col : str, optional
            The column name to use for the label from `ref_events`
            (the default is None)
        ref_times : List[Tuple[datetime, str]], optional
            Add one or more reference line/label using (the default is None)

        Returns
        -------
        LayoutDOM
            The bokeh plot figure.

        """
        return display_timeline_duration(
            data=self._df,
            group_by=group_by,
            time_column=time_column,
            end_time_column=end_time_column,
            **kwargs,
        )

    def process_tree(self, **kwargs) -> Tuple[figure, LayoutDOM]:
        """
        Build and plot a process tree.

        Parameters
        ----------
        schema : ProcSchema, optional
            The data schema to use for the data set, by default None
            (if None the schema is inferred)
        output_var : str, optional
            Output variable for selected items in the tree,
            by default None
        legend_col : str, optional
            The column used to color the tree items, by default None
        show_table: bool
            Set to True to show a data table, by default False.

        Other Parameters
        ----------------
        height : int, optional
            The height of the plot figure
            (the default is 700)
        width : int, optional
            The width of the plot figure (the default is 900)
        title : str, optional
            Title to display (the default is None)
        hide_legend : bool, optional
            Hide the legend box, even if legend_col is specified.
        pid_fmt : str, optional
            Display Process ID as 'dec' (decimal), 'hex' (hexadecimal),
            or 'guid' (string), default is 'hex'.

        Returns
        -------
        Tuple[figure, LayoutDOM]:
            figure - The main bokeh.plotting.figure
            Layout - Bokeh layout structure.

        """
        return build_and_show_process_tree(data=self._df, **kwargs)

    def matrix(self, **kwargs) -> LayoutDOM:
        """
        Plot data as an intersection matrix.

        Parameters
        ----------
        data : pd.DataFrame
            The data to plot.
        x : str
            Column to plot on the x (horizontal) axis
        x_col : str
            Alias for 'x'
        y : str
            Column to plot on the y (vertical) axis
        y_col : str
            Alias for 'y'
        title : str, optional
            Custom title, default is 'Intersection plot'
        value_col : str, optional
            Column from the DataFrame used to size the intersection points.
        dist_count : bool, optional
            Calculates a count of distinct values (from `value_col`) and uses
            this to size the intersection points.
            Requires `value_col` to be specified.
        log_size : bool, optional
            Takes the log of the size value before calculating the intersection
            display point size.
            Can be combined with `invert`.
        invert : bool, optional
            Takes the inverse of the size value as the basis for calculating
            the intersection display point size. This is useful for highlighting
            rare interactions.
            Can be combined with `log_size`.
        intersect : bool, optional
            Plots points of a fixed size, rather than using a sizing value. This
            is useful for just showing the presence/absence of an interaction.
        height : int, optional
            The plot height. Default is 700
        width : int
            The plot width. Default is 900
        color : str
            The color of the plotted points, default is "red"
        sort : Union[str, bool], optional
            Sorts the labels of both axes, default is None.
            Acceptable values are:
            'asc' (or string starting with 'asc') - Sort ascending
            'desc' (or string starting with 'asc') - Sort descending
            False or None (no sort)
            True  - Sort ascending
        sort_x : str, optional
            Sorts the labels of the x axis (takes precedence over `sort`),
            default is None.
            Acceptable values are:
            'asc' (or string starting with 'asc') - Sort ascending
            'desc' (or string starting with 'asc') - Sort descending
            False or None (no sort)
            True  - Sort ascending
        sort_y : str, optional
            Sorts the labels of the y axis (takes precedence over `sort`),
            default is None.
            Acceptable values are:
            'asc' (or string starting with 'asc') - Sort ascending
            'desc' (or string starting with 'asc') - Sort descending
            False or None (no sort)
            True  - Sort ascending
        hide : bool, optional
            Creates and returns but does not display the plot, default
            is False.
        font_size : int, optional
            Manually specify the font size for axis labels, the
            default is to automatically calculate a size based on the
            number of items in each axis.
        max_label_font_size : int, optional
            The maximum size, in points, of the X and Y labels, default is 11.


        Returns
        -------
        LayoutDOM
            The Bokeh plot

        """
        return plot_matrix(data=self._df, **kwargs)

    def incident_graph(
        self, timeline: bool = False, hide: bool = False, **kwargs
    ) -> LayoutDOM:
        """
        Plot an incident graph if the dataframe contains incidents or alerts.

        Parameters
        ----------
        timeline : bool, optional
            True to plot the entity timeline, by default False
        hide : bool, optional
            True to hide the plot, by default False

        Other Parameters
        ----------------
        node_size : int, optional
            Size of the nodes in pixels, by default 25
        font_size : int, optional
            Font size for node labels, by default 10
            Can be an integer (point size) or a string (e.g. "10pt")
        width : int, optional
            Width in pixels, by default 800
        height : int, optional
            Image height (the default is 800)
        scale : int, optional
            Position scale (the default is 2)

        Raises
        ------
        MsticpyUserError
            Raised if the dataframe does not contain incidents or alerts.

        """
        if any(elem not in self._df.columns for elem in req_alert_cols) and any(
            elem not in self._df.columns for elem in req_inc_cols
        ):
            raise MsticpyUserError("DataFrame must consist of Incidents or Alerts")
        graph = EntityGraph(self._df)
        return graph.plot(hide=hide, timeline=timeline, **kwargs)

    def folium_map(self, **kwargs):
        """
        Plot folium map from DataFrame.

        Parameters
        ----------
        ip_column : Optional[str], optional
            The name of the IP Address column, by default None
        lat_column : Optional[str], optional
            The name of the location 'latitude' column, by default None
        long_column : Optional[str], optional
            The name of the location 'longitude' column, by default None
        layer_column : Optional[str], optional
            The column to group markers into for displaying on different
            map layers, by default None
        icon_column : Optional[str], optional
            Optional column containing the name of the icon to use
            for the marker in this row, by default None
        icon_map : IconMapper, optional
            Mapping dictionary or function, by default None
            See Notes for more details.
        popup_columns : Optional[List[str]], optional
            List of columns to use for the popup text, by default None
        tooltip_columns : Optional[List[str]], optional
            List of columns to use for the tooltip text, by default None


        Other Parameters
        ----------------
        marker_cluster : bool, optional
            Use marker clustering, default is True.
        default_color : str, optional
            Default color for marker icons, by default "blue"
        title : str, optional
            Name of the layer (the default is 'layer1')
            (passed to FoliumMap constructor)
        zoom_start : int, optional
            The zoom level of the map (the default is 7)
            (passed to FoliumMap constructor)
        tiles : [type], optional
            Custom set of tiles or tile URL (the default is None)
            (passed to FoliumMap constructor)
        width : str, optional
            Map display width (the default is '100%')
            (passed to FoliumMap constructor)
        height : str, optional
            Map display height (the default is '100%')
            (passed to FoliumMap constructor)
        location : list, optional
            Location to center map on

        Returns
        -------
        folium.Map
            Folium Map object.

        Raises
        ------
        ValueError
            If neither `ip_column` nor `lat_column` and `long_column` are passed.
        LookupError
            If one of the passed columns does not exist in `data`

        Notes
        -----
        There are two ways of providing custom icon settings based on the
        the row of the input DataFrame.

        If `icon_map` is a dict it should contain keys that map to the
        value of `icon_col` and values that a dicts of valid
        folium Icon properties ("color", "icon_color", "icon", "angle", "prefix").
        The dict should include a "default" entry that will be used if the
        value in the DataFrame[icon_col] doesn't match any key.
        For example:

        .. code:: python

            icon_map = {
                "high": {
                    "color": "red",
                    "icon": "warning",
                },
                "medium": {
                    "color": "orange",
                    "icon": "triangle-exclamation",
                    "prefix": "fa",
                },
                "default": {
                    "color": "blue",
                    "icon": "info-sign",
                },
            }

        If icon_map is a function it should take a single str parameter
        (the item key) and return a dict of icon properties. It should
        return a default set of values if the key does not match a known
        key. The `icon_col` value for each row will be passed to this
        function and the return value used to populate the Icon arguments.

        For example:

        .. code::python

            def icon_mapper(icon_key):
                if icon_key.startswith("bad"):
                    return {
                        "color": "red",
                        "icon": "triangle-alert",
                    }
                ...
                else:
                    return {
                        "color": "blue",
                        "icon": "info-sign",
                    }

        FontAwesome icon (prefix "fa") names are available at https://fontawesome.com/
        GlyphIcons icons (prefix "glyphicon") are available at https://www.glyphicons.com/

        """
        return plot_map(data=self._df, **kwargs)

    # pylint: disable=too-many-arguments
    def network(
        self,
        source_col: str,
        target_col: str,
        title: str = "Data Graph",
        source_attrs: Optional[Iterable[str]] = None,
        target_attrs: Optional[Iterable[str]] = None,
        edge_attrs: Optional[Iterable[str]] = None,
        graph_type: GraphType = "graph",
        **kwargs,
    ):
        """
        Plot entity graph with Bokeh.

        Parameters
        ----------
        source_col : str
            Column for source nodes.
        target_col : str
            Column for target nodes.
        title : str
            Title for the plot, by default 'Data Graph'
        node_size : int, optional
            Size of the nodes in pixels, by default 25
        font_size : int, optional
            Font size for node labels, by default 10
            Can be an integer (point size) or a string (e.g. "10pt")
        width : int, optional
            Width in pixels, by default 800
        height : int, optional
            Image height (the default is 800)
        scale : int, optional
            Position scale (the default is 2)
        hide : bool, optional
            Don't show the plot, by default False. If True, just
            return the figure.
        source_attrs : Optional[List[str]], optional
            Optional list of source attributes to use as hover properties, by default None
        target_attrs : Optional[List[str]], optional
            Optional list of target attributes to use as hover properties, by default None
        edge_attrs : Optional[List[str]], optional
            Optional list of edge attributes to use as hover properties, by default None
        graph_type : str
            "graph" or "digraph" (for nx.DiGraph)

        Other Parameters
        ----------------
        source_color : str, optional
            The color of the source nodes, by default 'light-blue'
        target_color : str, optional
            The color of the source nodes, by default 'light-green'
        edge_color : str, optional
            The color of the edges, by default 'black'
        kwargs :
            Additional keyword arguments are passed to the networkx
            layout function.

        Returns
        -------
        bokeh.plotting.figure
            The network plot.

        """
        nx_graph = df_to_networkx(
            data=self._df,
            source_col=source_col,
            target_col=target_col,
            source_attrs=source_attrs,
            target_attrs=target_attrs,
            edge_attrs=edge_attrs,
            graph_type=graph_type,
        )
        return plot_nx_graph(
            nx_graph=nx_graph,
            title=title,
            source_attrs=source_attrs,
            target_attrs=target_attrs,
            edge_attrs=edge_attrs,
            **kwargs,
        )



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module for Log Analytics-related configuration."""
import contextlib
import json
import os
import re
from pathlib import Path
from typing import Any, Dict, Optional

import ipywidgets as widgets
from IPython.display import display

from .._version import VERSION
from .exceptions import MsticpyUserConfigError
from .pkg_config import get_config, refresh_config
from .utility import export, is_valid_uuid, md, md_warn

__version__ = VERSION
__author__ = "Ian Hellen"


_RESOURCES = [
    (
        "https://github.com/Azure/Azure-Sentinel-Notebooks/blob/"
        "master/ConfiguringNotebookEnvironment.ipynb"
    ),
    (
        "https://github.com/Azure/Azure-Sentinel-Notebooks/blob/"
        "master/A%20Getting%20Started%20Guide%20For%20Azure%20"
        "Sentinel%20ML%20Notebooks.ipynb"
    ),
]

_NO_CONFIG_WARN = [
    "Could not find Microsoft Sentinel settings in msticpyconfig.yaml or a config.json ",
    "(in the current directory or via a MSTICPYCONFIG variable.)",
    "We found the file '{config_file}' and will use this.",
    "We recommend using an explicit msticpyconfig.yaml specified using the",
    "MSTICPYCONFIG environment variable. See:",
    (
        "https://msticpy.readthedocs.io/en/latest/data_acquisition/DataProviders.html#"
        "connecting-to-an-azure-sentinel-workspace"
    ),
    "and",
    "https://msticpy.readthedocs.io/en/latest/getting_started/msticpyconfig.html",
    "for more details.",
    "Also see the notebooks:",
    *_RESOURCES,
]

_NO_CONFIG_ERR = [
    "Could not find msticpyconfig.yaml or config.json.",
    "The 'config.json' file is created when you launch notebooks from "
    "Microsoft Sentinel. If you have copied the notebook to another location "
    "or folder you will need to copy this configuration file.",
    "Alternatively, we recommend using an explicit msticpyconfig.yaml"
    "and adding your Workspace and Tenant IDs to that file.",
    "",
    "You can create a settings file using the following commands:",
    ">>> from msticpy.config import MpConfigEdit",
    ">>> MpConfigEdit()",
]

WIDGET_DEFAULTS = {
    "layout": widgets.Layout(width="95%"),
    "style": {"description_width": "150px"},
}


@export
class WorkspaceConfig:
    """Workspace configuration class."""

    # Constants
    TENANT_ID = "{{cookiecutter.tenant_id}}"
    SUBSCRIPTION_ID = "{{cookiecutter.subscription_id}}"
    RESOURCE_GROUP = "{{cookiecutter.resource_group}}"
    WORKSPACE_ID = "{{cookiecutter.workspace_id}}"
    WORKSPACE_NAME = "{{cookiecutter.workspace_name}}"

    CONF_TENANT_ID = "TenantId"
    CONF_WS_ID = "WorkspaceId"
    CONF_SUB_ID = "SubscriptionId"
    CONF_RES_GROUP = "ResourceGroup"
    CONF_WS_NAME = "WorkspaceName"
    CONF_ARGS = "Args"

    # Legacy key names / constants
    PKG_CONF_TENANT_KEY = "TenantId"
    PKG_CONF_WS_KEY = "WorkspaceId"
    PKG_CONF_SUB_KEY = "SubscriptionId"
    PKG_CONF_RES_GROUP_KEY = "ResourceGroup"
    PKG_CONF_NAME_KEY = "WorkspaceName"
    PKG_CONF_ARGS_KEY = "Args"

    CONF_WS_ID_KEY = "workspace_id"
    CONF_TENANT_ID_KEY = "tenant_id"
    CONF_SUB_ID_KEY = "subscription_id"
    CONF_RES_GROUP_KEY = "resource_group"
    CONF_WS_NAME_KEY = "workspace_name"
    CONF_ARGS_KEY = "args"

    _SETTINGS_TO_CONFIG_NAME_MAP = {
        CONF_TENANT_ID: CONF_TENANT_ID_KEY,
        CONF_WS_ID: CONF_WS_ID_KEY,
        CONF_SUB_ID: CONF_SUB_ID_KEY,
        CONF_RES_GROUP: CONF_RES_GROUP_KEY,
        CONF_WS_NAME: CONF_WS_NAME_KEY,
        CONF_ARGS: CONF_ARGS_KEY,
    }
    _CONFIG_TO_SETTINGS_NAME_MAP = {
        val: key for key, val in _SETTINGS_TO_CONFIG_NAME_MAP.items()
    }

    def __init__(
        self,
        workspace: Optional[str] = None,
        config_file: Optional[str] = None,
        interactive: bool = True,
        config: Optional[Dict[str, str]] = None,
    ):
        """
        Load current Azure Notebooks configuration for Log Analytics.

        Parameters
        ----------
        config_file : Optional[str], optional
            path to a configuration file,
            If not specified, the defaults is to use a configured msticpyconfig.yaml
            If this isn't configured, it will search for (first) a config.json
            and (second) a msticpyconfig.yaml in (first) the current directory
            and (second) the parent directory and subfolders.
        workspace : Optional[str]
            Workspace name (where multiple workspaces are configured),
            by default the Default workspace will be used.
        interactive : bool, optional
            If this is False, initializing the class will not raise an
            exception if no configuration is found. By default, True.
        config : Optional[Dict[str, str]],
            Workspace configuration as dictionary.

        """
        self._config: Dict[str, Any] = {}
        self._interactive = interactive
        self._config_file = config_file
        self.workspace_key = workspace or "Default"
        self.settings_key: Optional[str] = None

        # If config file specified, use that
        if config:
            self._config.update(config)
        elif config_file:
            self._config.update(self._read_config_values(config_file))
        else:
            self._determine_config_source(workspace)

    def __getattr__(self, attribute: str):
        """Return attribute from configuration."""
        with contextlib.suppress(KeyError):
            return self[attribute]
        raise AttributeError(
            f"{self.__class__.__name__} has no attribute '{attribute}'"
        )

    def __getitem__(self, key: str):
        """Allow property get using dictionary key syntax."""
        if key in self._SETTINGS_TO_CONFIG_NAME_MAP:
            return self._config.get(key)
        if key in self._CONFIG_TO_SETTINGS_NAME_MAP:
            return self._config.get(self._CONFIG_TO_SETTINGS_NAME_MAP[key])
        raise KeyError(f"{self.__class__.__name__} has no attribute '{key}'")

    def __setitem__(self, key: str, value: Any):
        """Allow property set using dictionary key syntax."""
        self._config[key] = value

    def __contains__(self, key: str):
        """Allow property in test."""
        # In operator overload
        return (
            key == "Type"
            or key in self._config
            or self._CONFIG_TO_SETTINGS_NAME_MAP.get(key) in self._config
            or key in self.__dict__
        )

    def __repr__(self):
        """Return contents of current config."""
        return self._config.__repr__()

    def get(self, key, default: Any = None):
        """Return key value or default."""
        try:
            return self[key]
        except KeyError:
            return default

    @property
    def config_loaded(self) -> bool:
        """
        Return True if workspace id and tenant id have values.

        Returns
        -------
        bool
            True if configuration loaded.

        """
        ws_value = self._config.get(self.CONF_WS_ID, None)
        ten_value = self._config.get(self.CONF_TENANT_ID, None)
        return is_valid_uuid(ws_value) and is_valid_uuid(ten_value)  # type: ignore

    @property
    def code_connect_str(self) -> str:
        """
        Return the Log Analytics connection string for dev code auth.

        Returns
        -------
        str
            Connection string

        """
        ten_id = self[self.CONF_TENANT_ID]
        ws_id = self[self.CONF_WS_ID]
        if not ten_id:
            raise KeyError(
                f"Configuration setting for {self.CONF_TENANT_ID} "
                + "could not be found."
            )
        if not ws_id:
            raise KeyError(
                f"Configuration setting for {self.CONF_WS_ID} " + "could not be found."
            )
        return f"loganalytics://code().tenant('{ten_id}').workspace('{ws_id}')"

    @property
    def mp_settings(self):
        """Return the equivalent MSTICPY settings dictionary."""
        return {
            self.CONF_WS_NAME: self._config.get(self.CONF_WS_NAME),
            self.CONF_SUB_ID: self._config.get(self.CONF_SUB_ID),
            self.CONF_WS_ID: self._config.get(self.CONF_WS_ID),
            self.CONF_TENANT_ID: self._config.get(self.CONF_TENANT_ID),
            self.CONF_RES_GROUP: self._config.get(self.CONF_RES_GROUP),
            self.CONF_ARGS: self._config.get(self.CONF_ARGS),
        }

    @property
    def args(self) -> Dict[str, str]:
        """Return any additional arguments."""
        return self._config.get(self.CONF_ARGS, {})

    @property
    def settings_path(self) -> Optional[str]:
        """Return the path to the settings in the MSTICPY config."""
        if self.settings_key:
            return f"AzureSentinel.Workspaces.{self.settings_key}"
        return None

    @property
    def settings(self) -> Dict[str, Any]:
        """Return the current settings dictionary."""
        return get_config(self.settings_path, {})

    @classmethod
    def from_settings(cls, settings: Dict[str, Any]) -> "WorkspaceConfig":
        """Create a WorkstationConfig from MSTICPY Workspace settings."""
        return cls(
            config={  # type: ignore
                cls.CONF_WS_NAME: settings.get(cls.CONF_WS_NAME),  # type: ignore
                cls.CONF_SUB_ID: settings.get(cls.CONF_SUB_ID),  # type: ignore
                cls.CONF_WS_ID: settings.get(cls.CONF_WS_ID),  # type: ignore
                cls.CONF_TENANT_ID: settings.get(cls.CONF_TENANT_ID),  # type: ignore
                cls.CONF_RES_GROUP: settings.get(cls.CONF_RES_GROUP),  # type: ignore
            }
        )

    @classmethod
    def from_connection_string(cls, connection_str: str) -> "WorkspaceConfig":
        """Create a WorkstationConfig from a connection string."""
        tenant_regex = r"""
        .*tenant\s?[=\(]\s?['\"]\{?
        (?P<tenant_id>[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12})
        \}?['\"].*"""
        workspace_regex = r"""
        .*workspace\s?[=\(]\s?['\"]\{?
        (?P<workspace_id>[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12})
        \}?['\"].*"""
        ws_name_regex = r".*alias\s?[=\(]\s?['\"]\{?(?P<workspace_name>\w+)['\"].*"

        tenant_id = workspace_id = workspace_name = None
        if match := re.match(tenant_regex, connection_str, re.IGNORECASE | re.VERBOSE):
            tenant_id = match.groupdict()["tenant_id"]
        else:
            raise ValueError("Could not find tenant ID in connection string.")
        if match := re.match(
            workspace_regex, connection_str, re.IGNORECASE | re.VERBOSE
        ):
            workspace_id = match.groupdict()["workspace_id"]
        else:
            raise ValueError("Could not find workspace ID in connection string.")
        if match := re.match(ws_name_regex, connection_str, re.IGNORECASE | re.VERBOSE):
            workspace_name = match.groupdict()["workspace_name"]
        return cls(
            config={
                cls.CONF_WS_ID: workspace_id,  # type: ignore[dict-item]
                cls.CONF_TENANT_ID: tenant_id,  # type: ignore[dict-item]
                cls.CONF_WS_NAME: workspace_name,  # type: ignore[dict-item]
            }
        )

    @classmethod
    def _read_config_values(cls, file_path: str) -> Dict[str, str]:
        """Read configuration file."""
        if not file_path:
            return {}
        with contextlib.suppress(json.JSONDecodeError):
            with open(file_path, "r", encoding="utf-8") as json_file:
                if json_file:
                    config_ws = json.load(json_file)
                    return {
                        cls._CONFIG_TO_SETTINGS_NAME_MAP[key]: value
                        for key, value in config_ws.items()
                        if key in cls._CONFIG_TO_SETTINGS_NAME_MAP
                    }
        return {}

    @classmethod
    def list_workspaces(cls) -> Dict:
        """
        Return list of available workspaces.

        Returns
        -------
        Dict
            Dictionary of workspaces with workspace and tenantIds.

        """
        ws_settings = get_config("AzureSentinel", {}).get("Workspaces")
        return (
            {
                ws_name: {
                    cls.CONF_WS_ID: ws.get(cls.CONF_WS_ID),
                    cls.CONF_TENANT_ID: ws.get(cls.CONF_TENANT_ID),
                }
                for ws_name, ws in ws_settings.items()
            }
            if ws_settings
            else {}
        )

    def prompt_for_ws(self):
        """Display an interactive prompt for Workspace details."""
        md_warn("No Microsoft Sentinel configuration found.")
        md(
            "Please enter the workspace ID and tenant Id"
            + " to allow connection to the Microsoft Sentinel workspace."
        )
        ws_id_wgt = widgets.Text(description="Workspace Id:", **WIDGET_DEFAULTS)
        ten_id_wgt = widgets.Text(description="Tenant Id:", **WIDGET_DEFAULTS)

        def update_ws(chg):
            self["workspace_id"] = chg.get("new")

        def update_tnt(chg):
            self["tenant_id"] = chg.get("new")

        ws_id_wgt.observe(update_ws, names="value")
        ten_id_wgt.observe(update_tnt, names="value")
        display(widgets.VBox([ws_id_wgt, ten_id_wgt]))
        md(
            (
                "You can avoid this prompt in future by following the"
                " guidance in the"
                " <a href='https://msticpy.readthedocs.io/en/latest/"
                "data_acquisition/DataProviders.html#"
                "connecting-to-an-azure-sentinel-workspace' "
                "target='_blank' rel='noopener noreferrer'>"
                "Connecting to a Microsoft Sentinel Workspace</a>"
            ),
        )

    def _determine_config_source(self, workspace):
        # First, try default MSTICPy config
        self._read_pkg_config_values(workspace_name=workspace)
        if self.config_loaded:
            return
        # Next, search for a config.json in the current directory
        if Path("./config.json").exists():
            self._config_file = "./config.json"
        else:
            self._config_file = self._search_for_file("**/config.json")
        if self._config_file:
            self._config.update(self._read_config_values(self._config_file))
            return

        # Finally, search for a msticpyconfig.yaml
        self._config_file = self._search_for_file("**/msticpyconfig.yaml")
        if self._config_file:
            os.environ["MSTICPYCONFIG"] = self._config_file
            refresh_config()
            self._read_pkg_config_values(workspace_name=workspace)
            return
        # Finally, throw an exception (unless non-interactive)
        if self._interactive:
            # If we've arrived here after searching current folder and parent
            # then we give up. (We create but don't raise an actual exception)
            display(
                MsticpyUserConfigError(
                    *_NO_CONFIG_ERR,
                    title="Workspace configuration missing.",
                    **{f"nb_{idx}_uri": res for idx, res in enumerate(_RESOURCES)},
                )
            )

    def _read_pkg_config_values(self, workspace_name: Optional[str] = None):
        """Try to find a usable config from the MSTICPy config file."""
        ws_settings = get_config("AzureSentinel", {}).get("Workspaces")  # type: ignore
        if not ws_settings:
            return
        selected_workspace: Dict[str, str] = {}
        if workspace_name:
            selected_workspace, self.settings_key = self._lookup_ws_name_and_id(
                workspace_name, ws_settings
            )
        elif "Default" in ws_settings:
            selected_workspace = ws_settings["Default"]
            self.settings_key = "Default"
        elif len(ws_settings) == 1:
            # If only one workspace, use that
            self.settings_key, selected_workspace = next(iter(ws_settings.items()))

        if selected_workspace:
            self._config = {}
            self._config.update(selected_workspace)

    def _lookup_ws_name_and_id(self, ws_name: str, ws_configs: dict):
        for name, ws_config in ws_configs.items():
            if ws_name.casefold() == name.casefold():
                return ws_config, name
            if ws_config.get(self.CONF_WS_ID, "").casefold() == ws_name.casefold():
                return ws_config, name
            if ws_config.get(self.CONF_WS_NAME, "").casefold() == ws_name.casefold():
                return ws_config, name
        return {}, None

    def _search_for_file(self, pattern: str) -> Optional[str]:
        config_file = None
        for start_path in (".", ".."):
            searched_configs = list(Path(start_path).glob(pattern))
            for found_file in searched_configs:
                if found_file.name == "msticpyconfig.yaml":
                    config_file = str(found_file)
                    break
                test_content = self._read_config_values(str(found_file))
                if "workspace_id" in test_content:
                    config_file = str(found_file)
                    break
            if config_file:
                break
        if config_file:
            # Warn that we're using a "found" file, not one in the current directory
            print("Warning")
            print("\n".join(_NO_CONFIG_WARN).format(config_file=config_file))
        return config_file



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module for common display functions."""
from typing import Any, List, Mapping, Tuple, Union

import IPython
import networkx as nx
import pandas as pd
from deprecated.sphinx import deprecated
from IPython.display import HTML, Javascript, display

from .._version import VERSION
from ..common.utility import export
from ..nbtools.security_alert import SecurityAlert
from .network_plot import plot_entity_graph

# pylint: disable=unused-import
from .timeline import display_timeline  # noqa: F401
from .timeline_values import display_timeline_values  # noqa: F401

# pylint: disable=unused-import

__version__ = VERSION
__author__ = "Ian Hellen"


@export
def display_alert(
    alert: Union[Mapping[str, Any], SecurityAlert], show_entities: bool = False
):
    """
    Display a Security Alert.

    Parameters
    ----------
    alert : Union[Mapping[str, Any], SecurityAlert]
        The alert to display as Mapping (e.g. pd.Series)
        or SecurityAlert
    show_entities : bool, optional
        Whether to display entities (the default is False)

    """
    output = format_alert(alert, show_entities)
    if not isinstance(output, tuple):
        output = [output]
    for disp_obj in output:
        display(disp_obj)


@export
def format_alert(
    alert: Union[Mapping[str, Any], SecurityAlert], show_entities: bool = False
) -> Union[IPython.display.HTML, Tuple[IPython.display.HTML, pd.DataFrame]]:
    """
    Get IPython displayable Security Alert.

    Parameters
    ----------
    alert : Union[Mapping[str, Any], SecurityAlert]
        The alert to display as Mapping (e.g. pd.Series)
        or SecurityAlert
    show_entities : bool, optional
        Whether to display entities (the default is False)

    Returns
    -------
    Union[IPython.display.HTML, Tuple[IPython.display.HTML, pd.DataFrame]]
        Single or tuple of displayable IPython objects

    Raises
    ------
    ValueError
        If the alert object is in an unknown format

    """
    if isinstance(alert, SecurityAlert):
        return HTML(alert.to_html(show_entities=show_entities))

    # Display subset of raw properties
    if isinstance(alert, pd.Series):
        return pd.DataFrame(alert)

    raise ValueError(f"Unrecognized alert object type {type(alert)}")


@export
def exec_remaining_cells():
    """Execute all cells below currently selected cell."""
    Javascript("Jupyter.notebook.execute_cells_below()")


@deprecated(
    reason=(
        "Matplotlib version 'draw_alert_entity_graph' "
        "no longer supported - use 'plot_entity_graph'"
    ),
    version="0.3.2",
)
@export
# pylint: disable=too-many-arguments
def draw_alert_entity_graph(
    nx_graph: nx.Graph,
    font_size: int = 12,
    height: int = 8,
    width: int = 8,
    margin: float = 0.3,
    scale: int = 1,
):
    """
    Draw networkX graph with matplotlib.

    Parameters
    ----------
    nx_graph : nx.Graph
        The NetworkX graph to draw
    font_size : int, optional
        base font size (the default is 12)
    height : int, optional
        Image height (the default is 8)
    width : int, optional
        Image width (the default is 8)
    margin : float, optional
        Image margin (the default is 0.3)
    scale : int, optional
        Position scale (the default is 1)

    """
    del margin
    return plot_entity_graph(
        entity_graph=nx_graph,
        font_size=font_size,
        height=height * 100,
        width=width * 100,
        scale=scale * 2,
    )


# Constants for Windows logon
_WIN_LOGON_TYPE_MAP = {
    0: "Unknown",
    2: "Interactive",
    3: "Network",
    4: "Batch",
    5: "Service",
    7: "Unlock",
    8: "NetworkCleartext",
    9: "NewCredentials",
    10: "RemoteInteractive",
    11: "CachedInteractive",
}
_WINDOWS_SID = {
    "S-1-0-0": "Null SID",
    "S-1-5-18": "LOCAL_SYSTEM",
    "S-1-5-19": "LOCAL_SERVICE",
    "S-1-5-20": "NETWORK_SERVICE",
}
_ADMINISTRATOR_SID = "500"
_GUEST_SID = "501"
_DOM_OR_MACHINE_SID = "S-1-5-21"


@export
def display_logon_data(
    logon_event: pd.DataFrame, alert: SecurityAlert = None, os_family: str = None
):
    """
    Display logon data for one or more events as HTML table.

    Parameters
    ----------
    logon_event : pd.DataFrame
        Dataframe containing one or more logon events
    alert : SecurityAlert, optional
        obtain os_family from the security alert
        (the default is None)
    os_family : str, optional
         explicitly specify os_family (Linux or Windows)
         (the default is None)

    Notes
    -----
    Currently only Windows Logon events.

    """
    display(format_logon(logon_event, alert, os_family))


@export
def format_logon(
    logon_event: Union[pd.DataFrame, pd.Series],
    alert: SecurityAlert = None,
    os_family: str = None,
) -> IPython.display.HTML:
    """
    Return logon data for one or more events as HTML table.

    Parameters
    ----------
    logon_event : Union[pd.DataFrame, pd.Series]
        Dataframe containing one or more logon events
        or Series containing a single logon event.
    alert : SecurityAlert, optional
        obtain os_family from the security alert
        (the default is None)
    os_family : str, optional
         explicitly specify os_family (Linux or Windows)
         (the default is None)

    Returns
    -------
    IPython.display.HTML :
        HTML display object

    """
    if not os_family:
        os_family = alert.os_family if alert else "Windows"

    logon_output = []
    if isinstance(logon_event, pd.DataFrame):
        for _, logon_row in logon_event.iterrows():
            logon_record = _fmt_single_row(logon_row, os_family)
            logon_output.append(
                "<tr class='cell_logon'><td class='cell_logon'>"
                + f"{'<br>'.join(logon_record)}</td></tr>"
            )
    elif isinstance(logon_event, pd.Series):
        logon_record = _fmt_single_row(logon_event, os_family)
        logon_output.append(
            "<tr class='cell_logon'><td class='cell_logon'>"
            + f"{'<br>'.join(logon_record)}</td></tr>"
        )

    t_style = """
        <style>
            .table_logon {border-collapse: collapse; width: 50%;}
            .cell_logon {border: 1px solid #ddd !important;
                text-align: left !important; padding: 15px !important;}
        </style>
        """
    return HTML(f"{t_style}<table class='table_logon'>{''.join(logon_output)}</table>")


def _fmt_single_row(logon_row: pd.Series, os_family: str) -> List[str]:
    """Format a pandas series logon record."""
    logon_record = [
        f"<b>Account: </b>{logon_row['TargetUserName']}",
        f"<b>Account Domain: </b>{logon_row['TargetDomainName']}",
        f"<b>Logon Time: </b>{logon_row['TimeGenerated']}",
    ]

    if os_family == "Windows":
        logon_type = logon_row["LogonType"]
        logon_desc_idx = logon_type
        if logon_type not in _WIN_LOGON_TYPE_MAP:
            logon_desc_idx = 0
        logon_record.append(
            f"<b>Logon type: </b>{logon_type}"
            + f"({_WIN_LOGON_TYPE_MAP[logon_desc_idx]})"
        )

    account_id = logon_row.TargetUserSid
    logon_record.append(f"<b>User Id/SID: </b>{account_id}")
    if os_family == "Windows":
        logon_record.extend(_format_sid_info(account_id))
    else:
        logon_record.append(f"<b>Audit user: </b>{logon_row['audit_user']}")

    session_id = logon_row["TargetLogonId"]
    sess_id = f"<b>Session id: </b>'{session_id}'"
    if session_id in ["0x3e7", "-1"]:
        sess_id += "System logon session"
    logon_record.append("")

    domain = logon_row["SubjectDomainName"]
    if not domain:
        subj_account = logon_row.SubjectUserName
    else:
        subj_account = f"{domain}/{logon_row.SubjectUserName}"
    logon_record.append(f"<b>Subject (source) account: </b>{subj_account}")

    logon_record.append(f"<b>Logon process: </b>{logon_row['LogonProcessName']}")
    logon_record.append(
        f"<b>Authentication: </b>{logon_row['AuthenticationPackageName']}"
    )
    logon_record.append(f"<b>Source IpAddress: </b>{logon_row['IpAddress']}")
    logon_record.append(f"<b>Source Host: </b>{logon_row['WorkstationName']}")
    logon_record.append(f"<b>Logon status: </b>{logon_row['Status']}")
    logon_record.append("")
    return logon_record


def _format_sid_info(sid):
    sid_info = []
    if not sid:
        return sid_info
    if sid in _WINDOWS_SID:
        sid_info.append(f"&nbsp;&nbsp;SID {sid} is {_WINDOWS_SID[sid]}")
    elif sid.endswith(_ADMINISTRATOR_SID):
        sid_info.append(f"&nbsp;&nbsp;SID {sid} is administrator")
    elif sid.endswith(_GUEST_SID):
        sid_info.append(f"&nbsp;&nbsp;SID {sid} is guest")
    if sid.startswith(_DOM_OR_MACHINE_SID):
        sid_info.append(f"&nbsp;&nbsp;SID {sid} is local machine or domain account")
    return sid_info



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Uses the Azure Python SDK to collect and return details related to Azure."""
import datetime
import logging
from typing import Any, Dict, List, Optional, Tuple

import attr
import numpy as np
import pandas as pd
from azure.common.exceptions import CloudError
from azure.core.exceptions import ClientAuthenticationError
from azure.mgmt.resource.subscriptions import SubscriptionClient

from ..._version import VERSION
from ...auth.azure_auth import (
    AzCredentials,
    AzureCloudConfig,
    az_connect,
    fallback_devicecode_creds,
    only_interactive_cred,
)
from ...common.exceptions import (
    MsticpyAzureConfigError,
    MsticpyImportExtraError,
    MsticpyNotConnectedError,
    MsticpyResourceError,
)

try:
    from azure.mgmt.network import NetworkManagementClient
    from azure.mgmt.resource import ResourceManagementClient

    try:
        # Try new version but keep backward compat with 1.0.1
        from azure.mgmt.monitor import MonitorManagementClient
    except ImportError:
        from azure.mgmt.monitor import MonitorClient as MonitorManagementClient  # type: ignore
    from azure.mgmt.compute import ComputeManagementClient
    from azure.mgmt.compute.models import VirtualMachineInstanceView
except ImportError as imp_err:
    raise MsticpyImportExtraError(
        "Cannot use this feature without these azure packages installed",
        "azure.mgmt.network",
        "azure.mgmt.resource",
        "azure.mgmt.monitor",
        "azure.mgmt.compute",
        title="Error importing azure module",
        extra="azure",
    ) from imp_err

__version__ = VERSION
__author__ = "Pete Bryan"

logger = logging.getLogger(__name__)

_CLIENT_MAPPING = {
    "sub_client": SubscriptionClient,
    "resource_client": ResourceManagementClient,
    "network_client": NetworkManagementClient,
    "monitoring_client": MonitorManagementClient,
    "compute_client": ComputeManagementClient,
}


# pylint: disable=too-few-public-methods, too-many-instance-attributes
# attr class doesn't need a method
@attr.s(auto_attribs=True)
class Items:
    """attr class to build resource details dictionary."""

    resource_id: Optional[str] = None
    name: Optional[str] = None
    resource_type: Optional[str] = None
    location: Optional[str] = None
    tags: Optional[Any] = None
    plan: Optional[Any] = None
    properties: Optional[Any] = None
    kind: Optional[str] = None
    managed_by: Optional[str] = None
    sku: Optional[str] = None
    identity: Optional[str] = None
    state: Any = None


@attr.s(auto_attribs=True)
class NsgItems:
    """attr class to build NSG rule dictionary."""

    rule_name: Optional[str] = None
    description: Optional[str] = None
    protocol: Optional[str] = None
    direction: Optional[str] = None
    src_ports: Optional[str] = None
    dst_ports: Optional[str] = None
    src_addrs: Optional[str] = None
    dst_addrs: Optional[str] = None
    action: Optional[str] = None


@attr.s(auto_attribs=True)
class InterfaceItems:
    """attr class to build network interface details dictionary."""

    interface_id: Optional[str] = None
    private_ip: Optional[str] = None
    private_ip_allocation: Optional[str] = None
    public_ip: Optional[str] = None
    public_ip_allocation: Optional[str] = None
    app_sec_group: Optional[List[Any]] = None
    subnet: Optional[str] = None
    subnet_nsg: Any = None
    subnet_route_table: Any = None


class AzureData:
    """Class for returning data on an Azure tenant."""

    def __init__(self, connect: bool = False, cloud: Optional[str] = None):
        """Initialize connector for Azure Python SDK."""
        self.az_cloud_config = AzureCloudConfig(cloud)
        self.connected = False
        self.credentials: Optional[AzCredentials] = None
        self.sub_client: Optional[SubscriptionClient] = None
        self.resource_client: Optional[ResourceManagementClient] = None
        self.network_client: Optional[NetworkManagementClient] = None
        self.monitoring_client: Optional[MonitorManagementClient] = None
        self.compute_client: Optional[ComputeManagementClient] = None
        self.cloud = cloud or self.az_cloud_config.cloud
        self.endpoints = self.az_cloud_config.endpoints
        logger.info("Initialized AzureData")
        if connect:
            self.connect()

    def connect(
        self,
        auth_methods: Optional[List] = None,
        tenant_id: Optional[str] = None,
        silent: bool = False,
        **kwargs,
    ):
        """
        Authenticate to the Azure SDK.

        Parameters
        ----------
        auth_methods : List, optional
            list of preferred authentication methods to use, by default None
        tenant_id : str, optional
            The tenant to authenticate against. If not supplied, the default
            tenant for the identity will be used.
        silent : bool, optional
            Set true to prevent output during auth process, by default False
        cloud : str, optional
            What Azure cloud to connect to.
            By default it will attempt to use the cloud setting from config file.
            If this is not set it will default to Azure Public Cloud
        **kwargs
            Additional keyword arguments to pass to the az_connect function.

        Raises
        ------
        CloudError
            If no valid credentials are found or if subscription client can't be created

        See Also
        --------
        msticpy.auth.azure_auth.az_connect : function to authenticate to Azure SDK

        """
        if kwargs.get("cloud"):
            logger.info("Setting cloud to %s", kwargs["cloud"])
            self.cloud = kwargs["cloud"]
            self.az_cloud_config = AzureCloudConfig(self.cloud)
        auth_methods = auth_methods or self.az_cloud_config.auth_methods
        tenant_id = tenant_id or self.az_cloud_config.tenant_id
        self.credentials = az_connect(
            auth_methods=auth_methods, tenant_id=tenant_id, silent=silent, **kwargs
        )
        if not self.credentials:
            raise CloudError("Could not obtain credentials.")
        self._check_client("sub_client")
        if only_interactive_cred(self.credentials.modern) and not silent:
            print("Check your default browser for interactive sign-in prompt.")

        self.sub_client = SubscriptionClient(
            credential=self.credentials.modern,
            base_url=self.az_cloud_config.resource_manager,
            credential_scopes=[self.az_cloud_config.token_uri],
        )
        if not self.sub_client:
            raise CloudError("Could not create a Subscription client.")
        logger.info("Connected to Azure Subscription Client")
        self.connected = True

    def get_subscriptions(self) -> pd.DataFrame:
        """
        Get details of all subscriptions within the tenant.

        Returns
        -------
        pd.DataFrame
            Details of the subscriptions present in the users tenant.

        Raises
        ------
        MsticpyNotConnectedError
            If .connect() has not been called

        """
        if self.connected is False:
            raise MsticpyNotConnectedError(
                "You need to connect to the service before using this function.",
                help_uri=MsticpyAzureConfigError.DEF_HELP_URI,
                title="Please call connect() before continuing.",
            )

        subscription_ids = []
        display_names = []
        states = []
        try:
            sub_list = list(self.sub_client.subscriptions.list())  # type: ignore
        except AttributeError:
            self._legacy_auth("sub_client")
            sub_list = list(self.sub_client.subscriptions.list())  # type: ignore

        for item in sub_list:  # type: ignore
            subscription_ids.append(item.subscription_id)  # type: ignore
            display_names.append(item.display_name)  # type: ignore
            states.append(str(item.state))  # type: ignore

        return pd.DataFrame(
            {
                "Subscription ID": subscription_ids,
                "Display Name": display_names,
                "State": states,
            }
        )

    def get_subscription_info(self, sub_id: str) -> dict:
        """
        Get information on a specific subscription.

        Parameters
        ----------
        sub_id : str
            The ID of the subscription to return details on.

        Returns
        -------
        dict
            Details on the selected subscription.

        Raises
        ------
        MsticpyNotConnectedError
            If .connect() has not been called.

        """
        if self.connected is False:
            raise MsticpyNotConnectedError(
                "You need to connect to the service before using this function.",
                help_uri=MsticpyAzureConfigError.DEF_HELP_URI,
                title="Please call connect() before continuing.",
            )
        try:
            sub = self.sub_client.subscriptions.get(sub_id)  # type: ignore
        except AttributeError:
            self._legacy_auth("sub_client")
            sub = self.sub_client.subscriptions.get(sub_id)  # type: ignore

        sub_loc = sub.subscription_policies.location_placement_id  # type: ignore
        return {
            "Subscription ID": sub.subscription_id,
            "Display Name": sub.display_name,
            "State": str(sub.state),
            "Subscription Location": sub_loc,
            "Subscription Quota": sub.subscription_policies.quota_id,  # type: ignore
            "Spending Limit": sub.subscription_policies.spending_limit,  # type: ignore
        }

    def list_sentinel_workspaces(self, sub_id: str) -> Dict[str, str]:
        """
        Return a list of Microsoft Sentinel workspaces in a Subscription.

        Parameters
        ----------
        sub_id : str
            The subscription ID to get a list of workspaces from.
            If not provided it will attempt to get sub_id from config files.

        Returns
        -------
        Dict
            A dictionary of workspace names and ids

        """
        print("Finding Microsoft Sentinel Workspaces...")
        res = self.get_resources(sub_id=sub_id)  # type: ignore
        # handle no results
        if isinstance(res, pd.DataFrame) and not res.empty:
            sentinel = res[
                (res["resource_type"] == "Microsoft.OperationsManagement/solutions")
                & (res["name"].str.startswith("SecurityInsights"))
            ]
            workspaces = []
            for wrkspace in sentinel["resource_id"]:
                res_details = self.get_resource_details(
                    sub_id=sub_id, resource_id=wrkspace  # type: ignore
                )
                workspaces.append(res_details["properties"]["workspaceResourceId"])

            workspaces_dict = {}
            for wrkspace in workspaces:
                name = wrkspace.split("/")[-1]
                workspaces_dict[name] = wrkspace
            return workspaces_dict

        print(f"No Microsoft Sentinel workspaces in {sub_id}")
        return {}

    # Get > List Aliases
    get_sentinel_workspaces = list_sentinel_workspaces

    def get_resources(  # noqa: MC0001
        self, sub_id: str, rgroup: Optional[str] = None, get_props: bool = False
    ) -> pd.DataFrame:
        """
        Return details on all resources in a subscription or Resource Group.

        Parameters
        ----------
        sub_id: str
            The subscription ID to get resources for
        rgroup: str (Optional)
            The name of a Resource Group to get resources for
        get_props: bool (Optional)
            Set to True if you want to get the full properties of every resource
            Warning this may be a slow process depending on the number of resources

        Returns
        -------
        pd.DataFrame
            A dataframe of resource details

        """
        # Check if connection and client required are already present
        if self.connected is False:
            raise MsticpyNotConnectedError(
                "You need to connect to the service before using this function.",
                help_uri=MsticpyAzureConfigError.DEF_HELP_URI,
                title="Please call connect() before continuing.",
            )

        self._check_client("resource_client", sub_id)

        resources = []  # type: List
        if rgroup is None:
            resources.extend(
                iter(self.resource_client.resources.list())  # type: ignore
            )
        else:
            resources.extend(
                iter(
                    self.resource_client.resources.list_by_resource_group(  # type: ignore
                        rgroup
                    )
                )
            )

        # Warn users about getting full properties for each resource
        if get_props:
            print("Collecting properties for every resource may take some time...")

        resource_items = []

        # Get properties for each resource
        for resource in resources:
            if get_props:
                if resource.type == "Microsoft.Compute/virtualMachines":
                    state = self._get_compute_state(
                        resource_id=resource.id, sub_id=sub_id
                    )
                else:
                    state = None
                try:
                    props = self.resource_client.resources.get_by_id(  # type: ignore
                        resource.id, "2019-08-01"
                    ).properties

                except CloudError:
                    props = self.resource_client.resources.get_by_id(  # type: ignore
                        resource.id, self._get_api(resource.id, sub_id=sub_id)
                    ).properties
            else:
                props = resource.properties
                state = None

            # Parse relevant resource attributes into a dataframe and return it
            resource_details = attr.asdict(
                Items(  # type: ignore
                    resource.id,
                    resource.name,
                    resource.type,
                    resource.location,
                    resource.tags,
                    resource.plan,
                    props,
                    resource.kind,
                    resource.managed_by,
                    resource.sku,
                    resource.identity,
                    state,
                )
            )
            resource_items.append(resource_details)

        return pd.DataFrame(resource_items)

    def get_resource_details(  # noqa: MC0001
        self,
        sub_id: str,
        resource_id: Optional[str] = None,
        resource_details: Optional[dict] = None,
    ) -> dict:
        """
        Return the details of a specific Azure resource.

        Parameters
        ----------
        resource_id: str, optional
            The ID of the resource to get details on
        resource_details: dict, optional
            If ID is unknown provide the following details:
                -resource_group_name
                -resource_provider_namespace
                -resource_type
                -resource_name
                -parent_resource_path
        sub_id: str
            The ID of the subscription to get resources from

        Returns
        -------
        resource_details: dict
            The details of the requested resource

        """
        # Check if connection and client required are already present
        if self.connected is False:
            raise MsticpyNotConnectedError(
                "You need to connect to the service before using this function.",
                help_uri=MsticpyAzureConfigError.DEF_HELP_URI,
                title="Please call connect() before continuing.",
            )
        self._check_client("resource_client", sub_id)

        # If a resource id is provided use get_by_id to get details
        if resource_id is not None:
            try:
                resource = self.resource_client.resources.get_by_id(  # type: ignore
                    resource_id, api_version=self._get_api(resource_id, sub_id=sub_id)
                )
            except AttributeError:
                self._legacy_auth("resource_client", sub_id)
                resource = self.resource_client.resources.get_by_id(  # type: ignore
                    resource_id, api_version=self._get_api(resource_id, sub_id=sub_id)
                )
            if resource.type == "Microsoft.Compute/virtualMachines":
                state = self._get_compute_state(resource_id=resource_id, sub_id=sub_id)
            else:
                state = None
        # If resource details are provided use get to get details
        elif resource_details is not None:
            try:
                resource = self.resource_client.resources.get(  # type: ignore
                    resource_details["resource_group_name"],
                    resource_details["resource_provider_namespace"],
                    resource_details["parent_resource_path"],
                    resource_details["resource_type"],
                    resource_details["resource_name"],
                    api_version=self._get_api(
                        resource_provider=(
                            resource_details["resource_provider_namespace"]
                            + "/"
                            + resource_details["resource_type"]
                        ),
                        sub_id=sub_id,
                    ),
                )
            except AttributeError:
                self._legacy_auth("resource_client", sub_id)
                resource = self.resource_client.resources.get(  # type: ignore
                    resource_details["resource_group_name"],
                    resource_details["resource_provider_namespace"],
                    resource_details["parent_resource_path"],
                    resource_details["resource_type"],
                    resource_details["resource_name"],
                    api_version=self._get_api(
                        resource_provider=(
                            resource_details["resource_provider_namespace"]
                            + "/"
                            + resource_details["resource_type"]
                        ),
                        sub_id=sub_id,
                    ),
                )
            state = None
        else:
            raise ValueError("Please provide either a resource ID or resource details")

        # Parse relevent details into a dictionary to return
        resource_details = attr.asdict(
            Items(  # type: ignore
                resource.id,
                resource.name,
                resource.type,
                resource.location,
                resource.tags,
                resource.plan,
                resource.properties,
                resource.kind,
                resource.managed_by,
                resource.sku,
                resource.identity,
                state,
            )
        )

        return resource_details

    def _get_api(  # noqa: MC0001
        self,
        resource_id: Optional[str] = None,
        sub_id: Optional[str] = None,
        resource_provider: Optional[str] = None,
    ) -> str:
        """
        Return the latest avaliable API version for the resource.

        Parameters
        ----------
        resource_id: str, optional
            The ID of the resources to get an API version for
        sub_id: str, optional
            The ID of the subscription to get details from
        resource_provider: str, optional
            The resource provider namespace and service to get an API version for

        Returns
        -------
        api_ver: str
            The latest available non-preview API version

        """
        # Check if connection and client required are already present
        if self.connected is False:
            raise MsticpyNotConnectedError(
                "You need to connect to the service before using this function.",
                help_uri=MsticpyAzureConfigError.DEF_HELP_URI,
                title="Please call connect() before continuing.",
            )

        self._check_client("resource_client", sub_id)  # type: ignore

        # Normalize elements depending on user input type
        if resource_id is not None:
            try:
                namespace = resource_id.split("/")[6]
                service = resource_id.split("/")[7]
            except IndexError as idx_err:
                raise MsticpyResourceError(
                    "Provided Resource ID isn't in the correct format.",
                    "It should look like:",
                    "/subscriptions/SUB_ID/resourceGroups/RESOURCE_GROUP/"
                    + "providers/NAMESPACE/SERVICE_NAME/RESOURCE_NAME ",
                ) from idx_err

        elif resource_provider is not None:
            try:
                namespace = resource_provider.split("/")[0]
                service = resource_provider.split("/")[1]
            except IndexError as idx_err:
                raise MsticpyResourceError(
                    "Provided Resource Provider isn't in the correct format.",
                    "It should look like: NAMESPACE/SERVICE_NAME",
                ) from idx_err
        else:
            raise ValueError(
                "Please provide an resource ID or resource provider namespace"
            )

        # Get list of API versions for the service
        try:
            provider = self.resource_client.providers.get(namespace)  # type: ignore
        except AttributeError:
            self._legacy_auth("resource_client", sub_id)  # type: ignore
            provider = self.resource_client.providers.get(namespace)  # type: ignore

        resource_types = next(
            (
                t
                for t in provider.resource_types  # type: ignore
                if t.resource_type == service
            ),
            None,
        )

        # Get first API version that isn't in preview
        if not resource_types:
            raise MsticpyResourceError("Resource provider not found")

        api_version = [
            v
            for v in resource_types.api_versions  # type: ignore
            if "preview" not in v.lower()
        ]
        if api_version is None or not api_version:
            api_ver = resource_types.api_versions[0]  # type: ignore
        else:
            api_ver = api_version[0]
        return str(api_ver)

    def get_network_details(
        self, network_id: str, sub_id: str
    ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        Return details related to an Azure network interface and associated NSG.

        Parameters
        ----------
        network_id: str
            The ID of the network interface to return details on
        sub_id: str
            The subscription ID that the network interface is part of

        Returns
        -------
        details: dict
            A dictionary of items related to the network interface

        """
        # Check if connection and client required are already present
        if self.connected is False:
            raise MsticpyNotConnectedError(
                "You need to connect to the service before using this function.",
                help_uri=MsticpyAzureConfigError.DEF_HELP_URI,
                title="Please call connect() before continuing.",
            )

        self._check_client("network_client", sub_id)

        # Get interface details and parse relevant elements into a dataframe
        try:
            details = self.network_client.network_interfaces.get(  # type: ignore
                network_id.split("/")[4], network_id.split("/")[8]
            )
        except AttributeError:
            self._legacy_auth("network_client", sub_id)
            details = self.network_client.network_interfaces.get(  # type: ignore
                network_id.split("/")[4], network_id.split("/")[8]
            )

        ips = []
        for ip_addr in details.ip_configurations:  # type: ignore
            ip_details = attr.asdict(
                InterfaceItems(  # type: ignore
                    id=network_id,
                    private_ip=ip_addr.private_ip_address,
                    private_ip_allocation=str(ip_addr.private_ip_allocation_method),
                    public_ip=(
                        ip_addr.public_ip_address.ip_address
                        if ip_addr.public_ip_address
                        else None
                    ),
                    public_ip_allocation=(
                        ip_addr.public_ip_address.public_ip_allocation_method
                        if ip_addr.public_ip_address
                        else None
                    ),
                    app_sec_group=ip_addr.application_security_groups,  # type: ignore
                    subnet=ip_addr.subnet.name,  # type: ignore
                    subnet_nsg=ip_addr.subnet.network_security_group,  # type: ignore
                    subnet_route_table=ip_addr.subnet.route_table,  # type: ignore
                )
            )
            ips.append(ip_details)

        ip_df = pd.DataFrame(ips)

        nsg_df = pd.DataFrame()
        if details.network_security_group is not None:
            # Get NSG details and parse relevant elements into a dataframe
            nsg_details = self.network_client.network_security_groups.get(  # type: ignore
                details.network_security_group.id.split("/")[4],  # type: ignore
                details.network_security_group.id.split("/")[8],  # type: ignore
            )
            nsg_rules = []
            for nsg in nsg_details.default_security_rules:  # type: ignore
                rules = attr.asdict(
                    NsgItems(  # type: ignore
                        rule_name=nsg.name,
                        description=nsg.description,
                        protocol=str(nsg.protocol),
                        direction=str(nsg.direction),
                        src_ports=nsg.source_port_range,
                        dst_ports=nsg.destination_port_range,
                        src_addrs=nsg.source_address_prefix,
                        dst_addrs=nsg.destination_address_prefix,
                        action=str(nsg.access),
                    )
                )
                nsg_rules.append(rules)

            nsg_df = pd.DataFrame(nsg_rules)

        return ip_df, nsg_df

    def get_metrics(  # pylint: disable=too-many-locals
        self,
        metrics: str,
        resource_id: str,
        sub_id: str,
        sample_time: str = "hour",
        start_time: int = 30,
    ) -> Dict[str, pd.DataFrame]:
        """
        Return specified metrics on Azure Resource.

        Parameters
        ----------
        metrics: str
            A string list of metrics you wish to collect
            (https://docs.microsoft.com/en-us/azure/azure-monitor/platform/metrics-supported)
        resource_id: str
            The resource ID of the resource to collet the metrics from
        sub_id: str
            The subscription ID that the resource is part of
        sample_time: str (Optional)
            You can select to collect the metrics every hour of minute - default is hour
            Accepted inputs = 'hour' or 'minute'
        start_time: int (Optional)
            The number of days prior to today to collect metrics for, default is 30

        Returns
        -------
        results: dict
            A Dictionary of DataFrames containing the metrics details

        """
        if sample_time.casefold().startswith("h"):
            interval = "PT1H"
        elif sample_time.casefold().startswith("m"):
            interval = "PT1M"
        else:
            raise ValueError(
                "invalid value for sample_time - specify 'hour', or 'minute'"
            )

        # Check if connection and client required are already present
        if self.connected is False:
            raise MsticpyNotConnectedError(
                "You need to connect to the service before using this function.",
                help_uri=MsticpyAzureConfigError.DEF_HELP_URI,
                title="Please call connect() before continuing.",
            )

        self._check_client("monitoring_client", sub_id)

        # Get metrics in one hour chunks for the last 30 days
        start = datetime.datetime.now().date()
        end = start - datetime.timedelta(days=start_time)

        try:
            mon_details = self.monitoring_client.metrics.list(  # type: ignore
                resource_id,
                timespan=f"{end}/{start}",
                interval=interval,  # type: ignore
                metricnames=f"{metrics}",
                aggregation="Total",
            )
        except AttributeError:
            self._legacy_auth("monitoring_client", sub_id)
            mon_details = self.monitoring_client.metrics.list(  # type: ignore
                resource_id,
                timespan=f"{end}/{start}",
                interval=interval,  # type: ignore
                metricnames=f"{metrics}",
                aggregation="Total",
            )
        results = {}
        # Create a dict of all the results returned
        for metric in mon_details.value:  # type: ignore
            times = []
            output = []
            for time in metric.timeseries:
                for data in time.data:
                    times.append(data.time_stamp)
                    output.append(data.total)
            details = pd.DataFrame({"Time": times, "Data": output})
            details.replace(np.nan, 0, inplace=True)
            results[metric.name.value] = details
        return results

    # pylint: enable=too-many-locals, too-many-arguments

    def _get_compute_state(
        self, resource_id: str, sub_id: str
    ) -> VirtualMachineInstanceView:
        """
        Return the details on a Virtual Machine instance.

        Parameters
        ----------
        resource_id: str
            The Resource ID of the Virtual Machine
        sub_id: str
            The Subscription the Virtual Machine is part of

        Returns
        -------
        instance_details: VirtualMachineInstanceView
            The details of the Virtual Machine

        """
        if self.connected is False:
            raise MsticpyNotConnectedError(
                "You need to connect to the service before using this function.",
                help_uri=MsticpyAzureConfigError.DEF_HELP_URI,
                title="Please call connect() before continuing.",
            )

        self._check_client("compute_client", sub_id)

        # Parse the Resource ID to extract Resource Group and Resource Name
        r_details = resource_id.split("/")
        r_group = r_details[r_details.index("resourceGroups") + 1]
        name = r_details[r_details.index("virtualMachines") + 1]

        # Get VM instance details and return them
        try:
            instance_details = self.compute_client.virtual_machines.instance_view(  # type: ignore
                r_group, name
            )
        except AttributeError:
            self._legacy_auth("compute_client", sub_id)
            instance_details = self.compute_client.virtual_machines.instance_view(  # type: ignore
                r_group, name
            )

        return instance_details  # type: ignore

    def _check_client(self, client_name: str, sub_id: Optional[str] = None):
        """
        Check required client is present, if not create it.

        Parameters
        ----------
        client_name : str
            The name of the client to be checked.
        sub_id : str, optional
            The subscription ID for the client to connect to, by default None

        """
        if getattr(self, client_name) is None:
            client = _CLIENT_MAPPING[client_name]
            if sub_id is None:
                setattr(
                    self,
                    client_name,
                    client(
                        self.credentials.modern,  # type: ignore
                        base_url=self.az_cloud_config.resource_manager,
                        credential_scopes=[self.az_cloud_config.token_uri],
                    ),
                )
            else:
                setattr(
                    self,
                    client_name,
                    client(
                        self.credentials.modern,  # type: ignore
                        sub_id,
                        base_url=self.az_cloud_config.resource_manager,
                        credential_scopes=[self.az_cloud_config.token_uri],
                    ),
                )

        if getattr(self, client_name) is None:
            raise CloudError("Could not create client")

    def _legacy_auth(self, client_name: str, sub_id: Optional[str] = None):
        """
        Create client with v1 authentication token.

        Parameters
        ----------
        client_name : str
            The name of the client to be checked.
        sub_id : str, optional
            The subscription ID for the client to connect to, by default None

        """
        client = _CLIENT_MAPPING[client_name]
        if sub_id is None:
            setattr(
                self,
                client_name,
                client(
                    self.credentials.legacy,  # type: ignore
                    base_url=self.az_cloud_config.resource_manager,
                    credential_scopes=[self.az_cloud_config.token_uri],
                ),
            )
        else:
            setattr(
                self,
                client_name,
                client(
                    self.credentials.legacy,  # type: ignore
                    sub_id,
                    base_url=self.az_cloud_config.resource_manager,
                    credential_scopes=[self.az_cloud_config.token_uri],
                ),
            )


def get_api_headers(token: str) -> Dict:
    """
    Return authorization header with current token.

    Parameters
    ----------
    token : str
        Azure auth token.

    Returns
    -------
    Dict
        A dictionary of headers to be used in API calls.

    """
    return {
        "Authorization": f"Bearer {token}",
        "Content-Type": "application/json",
    }


def get_token(
    credential: AzCredentials,
    tenant_id: Optional[str] = None,
    cloud: Optional[str] = None,
) -> str:
    """
    Extract token from a azure.identity object.

    Parameters
    ----------
    credential : AzCredentials
        Azure OAuth credentials.
    tenant_id : str, optional
        The tenant to connect to if not the users home tenant.
    cloud: str, optional
        The Azure cloud to connect to.

    Returns
    -------
    str
        A token to be used in API calls.

    """
    az_cloud_config = AzureCloudConfig(cloud)
    tenant_id = tenant_id or az_cloud_config.tenant_id
    if tenant_id:
        try:
            token = credential.modern.get_token(AzureCloudConfig().token_uri)
        except ClientAuthenticationError:
            credential = fallback_devicecode_creds(cloud=cloud)
            token = credential.modern.get_token(AzureCloudConfig().token_uri)
    else:
        try:
            token = credential.modern.get_token(
                AzureCloudConfig().token_uri, tenant_id=tenant_id
            )
        except ClientAuthenticationError:
            credential = fallback_devicecode_creds(cloud=cloud, tenant_id=tenant_id)
            token = credential.modern.get_token(
                AzureCloudConfig().token_uri, tenant_id=tenant_id
            )

    return token.token
